import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    default_action: make_dataclass(
        "DefaultAction",
        [
            (
                "Block",
                make_dataclass(
                    "BlockAction",
                    [
                        (
                            "CustomResponse",
                            make_dataclass(
                                "CustomResponse",
                                [
                                    ("ResponseCode", int),
                                    ("CustomResponseBodyKey", str, field(default=None)),
                                    (
                                        "ResponseHeaders",
                                        List[
                                            make_dataclass(
                                                "CustomHTTPHeader",
                                                [("Name", str), ("Value", str)],
                                            )
                                        ],
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
            (
                "Allow",
                make_dataclass(
                    "AllowAction",
                    [
                        (
                            "CustomRequestHandling",
                            make_dataclass(
                                "CustomRequestHandling",
                                [
                                    (
                                        "InsertHeaders",
                                        List[
                                            make_dataclass(
                                                "CustomHTTPHeader",
                                                [("Name", str), ("Value", str)],
                                            )
                                        ],
                                        field(default=None),
                                    )
                                ],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
        ],
    ),
    visibility_config: make_dataclass(
        "VisibilityConfig",
        [
            ("SampledRequestsEnabled", bool),
            ("CloudWatchMetricsEnabled", bool),
            ("MetricName", str),
        ],
    ),
    resource_id: str = None,
    description: str = None,
    rules: List[
        make_dataclass(
            "Rule",
            [
                ("Name", str),
                ("Priority", int),
                (
                    "Statement",
                    make_dataclass(
                        "Statement",
                        [
                            (
                                "ByteMatchStatement",
                                make_dataclass(
                                    "ByteMatchStatement",
                                    [
                                        ("SearchString", bytes),
                                        (
                                            "FieldToMatch",
                                            make_dataclass(
                                                "FieldToMatch",
                                                [
                                                    (
                                                        "SingleHeader",
                                                        make_dataclass(
                                                            "SingleHeader",
                                                            [("Name", str)],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "SingleQueryArgument",
                                                        make_dataclass(
                                                            "SingleQueryArgument",
                                                            [("Name", str)],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "AllQueryArguments",
                                                        Dict,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "UriPath",
                                                        Dict,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "QueryString",
                                                        Dict,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "Body",
                                                        make_dataclass(
                                                            "Body",
                                                            [
                                                                (
                                                                    "OversizeHandling",
                                                                    str,
                                                                    field(default=None),
                                                                )
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "Method",
                                                        Dict,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "JsonBody",
                                                        make_dataclass(
                                                            "JsonBody",
                                                            [
                                                                (
                                                                    "MatchPattern",
                                                                    make_dataclass(
                                                                        "JsonMatchPattern",
                                                                        [
                                                                            (
                                                                                "All",
                                                                                Dict,
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                            (
                                                                                "IncludedPaths",
                                                                                List[
                                                                                    str
                                                                                ],
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                        ],
                                                                    ),
                                                                ),
                                                                ("MatchScope", str),
                                                                (
                                                                    "InvalidFallbackBehavior",
                                                                    str,
                                                                    field(default=None),
                                                                ),
                                                                (
                                                                    "OversizeHandling",
                                                                    str,
                                                                    field(default=None),
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "Headers",
                                                        make_dataclass(
                                                            "Headers",
                                                            [
                                                                (
                                                                    "MatchPattern",
                                                                    make_dataclass(
                                                                        "HeaderMatchPattern",
                                                                        [
                                                                            (
                                                                                "All",
                                                                                Dict,
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                            (
                                                                                "IncludedHeaders",
                                                                                List[
                                                                                    str
                                                                                ],
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                            (
                                                                                "ExcludedHeaders",
                                                                                List[
                                                                                    str
                                                                                ],
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                        ],
                                                                    ),
                                                                ),
                                                                ("MatchScope", str),
                                                                (
                                                                    "OversizeHandling",
                                                                    str,
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "Cookies",
                                                        make_dataclass(
                                                            "Cookies",
                                                            [
                                                                (
                                                                    "MatchPattern",
                                                                    make_dataclass(
                                                                        "CookieMatchPattern",
                                                                        [
                                                                            (
                                                                                "All",
                                                                                Dict,
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                            (
                                                                                "IncludedCookies",
                                                                                List[
                                                                                    str
                                                                                ],
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                            (
                                                                                "ExcludedCookies",
                                                                                List[
                                                                                    str
                                                                                ],
                                                                                field(
                                                                                    default=None
                                                                                ),
                                                                            ),
                                                                        ],
                                                                    ),
                                                                ),
                                                                ("MatchScope", str),
                                                                (
                                                                    "OversizeHandling",
                                                                    str,
                                                                ),
                                                            ],
                                                        ),
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                        ),
                                        (
                                            "TextTransformations",
                                            List[
                                                make_dataclass(
                                                    "TextTransformation",
                                                    [("Priority", int), ("Type", str)],
                                                )
                                            ],
                                        ),
                                        ("PositionalConstraint", str),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "SqliMatchStatement",
                                make_dataclass(
                                    "SqliMatchStatement",
                                    [
                                        ("FieldToMatch", "FieldToMatch"),
                                        (
                                            "TextTransformations",
                                            List["TextTransformation"],
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "XssMatchStatement",
                                make_dataclass(
                                    "XssMatchStatement",
                                    [
                                        ("FieldToMatch", "FieldToMatch"),
                                        (
                                            "TextTransformations",
                                            List["TextTransformation"],
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "SizeConstraintStatement",
                                make_dataclass(
                                    "SizeConstraintStatement",
                                    [
                                        ("FieldToMatch", "FieldToMatch"),
                                        ("ComparisonOperator", str),
                                        ("Size", int),
                                        (
                                            "TextTransformations",
                                            List["TextTransformation"],
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "GeoMatchStatement",
                                make_dataclass(
                                    "GeoMatchStatement",
                                    [
                                        (
                                            "CountryCodes",
                                            List[str],
                                            field(default=None),
                                        ),
                                        (
                                            "ForwardedIPConfig",
                                            make_dataclass(
                                                "ForwardedIPConfig",
                                                [
                                                    ("HeaderName", str),
                                                    ("FallbackBehavior", str),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "RuleGroupReferenceStatement",
                                make_dataclass(
                                    "RuleGroupReferenceStatement",
                                    [
                                        ("ARN", str),
                                        (
                                            "ExcludedRules",
                                            List[
                                                make_dataclass(
                                                    "ExcludedRule", [("Name", str)]
                                                )
                                            ],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "IPSetReferenceStatement",
                                make_dataclass(
                                    "IPSetReferenceStatement",
                                    [
                                        ("ARN", str),
                                        (
                                            "IPSetForwardedIPConfig",
                                            make_dataclass(
                                                "IPSetForwardedIPConfig",
                                                [
                                                    ("HeaderName", str),
                                                    ("FallbackBehavior", str),
                                                    ("Position", str),
                                                ],
                                            ),
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "RegexPatternSetReferenceStatement",
                                make_dataclass(
                                    "RegexPatternSetReferenceStatement",
                                    [
                                        ("ARN", str),
                                        ("FieldToMatch", "FieldToMatch"),
                                        (
                                            "TextTransformations",
                                            List["TextTransformation"],
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "RateBasedStatement",
                                make_dataclass(
                                    "RateBasedStatement",
                                    [
                                        ("Limit", int),
                                        ("AggregateKeyType", str),
                                        (
                                            "ScopeDownStatement",
                                            "Statement",
                                            field(default=None),
                                        ),
                                        (
                                            "ForwardedIPConfig",
                                            "ForwardedIPConfig",
                                            field(default=None),
                                        ),
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "AndStatement",
                                make_dataclass(
                                    "AndStatement",
                                    [
                                        (
                                            "Statements",
                                            List[
                                                make_dataclass(
                                                    "Statement",
                                                    [
                                                        (
                                                            "ByteMatchStatement",
                                                            "ByteMatchStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "SqliMatchStatement",
                                                            "SqliMatchStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "XssMatchStatement",
                                                            "XssMatchStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "SizeConstraintStatement",
                                                            "SizeConstraintStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "GeoMatchStatement",
                                                            "GeoMatchStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "RuleGroupReferenceStatement",
                                                            "RuleGroupReferenceStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "IPSetReferenceStatement",
                                                            "IPSetReferenceStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "RegexPatternSetReferenceStatement",
                                                            "RegexPatternSetReferenceStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "RateBasedStatement",
                                                            "RateBasedStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "AndStatement",
                                                            "AndStatement",
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "OrStatement",
                                                            make_dataclass(
                                                                "OrStatement",
                                                                [
                                                                    (
                                                                        "Statements",
                                                                        List[
                                                                            "Statement"
                                                                        ],
                                                                    )
                                                                ],
                                                            ),
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "NotStatement",
                                                            make_dataclass(
                                                                "NotStatement",
                                                                [
                                                                    (
                                                                        "Statement",
                                                                        "Statement",
                                                                    )
                                                                ],
                                                            ),
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "ManagedRuleGroupStatement",
                                                            make_dataclass(
                                                                "ManagedRuleGroupStatement",
                                                                [
                                                                    ("VendorName", str),
                                                                    ("Name", str),
                                                                    (
                                                                        "Version",
                                                                        str,
                                                                        field(
                                                                            default=None
                                                                        ),
                                                                    ),
                                                                    (
                                                                        "ExcludedRules",
                                                                        List[
                                                                            "ExcludedRule"
                                                                        ],
                                                                        field(
                                                                            default=None
                                                                        ),
                                                                    ),
                                                                    (
                                                                        "ScopeDownStatement",
                                                                        "Statement",
                                                                        field(
                                                                            default=None
                                                                        ),
                                                                    ),
                                                                    (
                                                                        "ManagedRuleGroupConfigs",
                                                                        List[
                                                                            make_dataclass(
                                                                                "ManagedRuleGroupConfig",
                                                                                [
                                                                                    (
                                                                                        "LoginPath",
                                                                                        str,
                                                                                        field(
                                                                                            default=None
                                                                                        ),
                                                                                    ),
                                                                                    (
                                                                                        "PayloadType",
                                                                                        str,
                                                                                        field(
                                                                                            default=None
                                                                                        ),
                                                                                    ),
                                                                                    (
                                                                                        "UsernameField",
                                                                                        make_dataclass(
                                                                                            "UsernameField",
                                                                                            [
                                                                                                (
                                                                                                    "Identifier",
                                                                                                    str,
                                                                                                )
                                                                                            ],
                                                                                        ),
                                                                                        field(
                                                                                            default=None
                                                                                        ),
                                                                                    ),
                                                                                    (
                                                                                        "PasswordField",
                                                                                        make_dataclass(
                                                                                            "PasswordField",
                                                                                            [
                                                                                                (
                                                                                                    "Identifier",
                                                                                                    str,
                                                                                                )
                                                                                            ],
                                                                                        ),
                                                                                        field(
                                                                                            default=None
                                                                                        ),
                                                                                    ),
                                                                                ],
                                                                            )
                                                                        ],
                                                                        field(
                                                                            default=None
                                                                        ),
                                                                    ),
                                                                ],
                                                            ),
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "LabelMatchStatement",
                                                            make_dataclass(
                                                                "LabelMatchStatement",
                                                                [
                                                                    ("Scope", str),
                                                                    ("Key", str),
                                                                ],
                                                            ),
                                                            field(default=None),
                                                        ),
                                                        (
                                                            "RegexMatchStatement",
                                                            make_dataclass(
                                                                "RegexMatchStatement",
                                                                [
                                                                    (
                                                                        "RegexString",
                                                                        str,
                                                                    ),
                                                                    (
                                                                        "FieldToMatch",
                                                                        "FieldToMatch",
                                                                    ),
                                                                    (
                                                                        "TextTransformations",
                                                                        List[
                                                                            "TextTransformation"
                                                                        ],
                                                                    ),
                                                                ],
                                                            ),
                                                            field(default=None),
                                                        ),
                                                    ],
                                                )
                                            ],
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("OrStatement", "OrStatement", field(default=None)),
                            ("NotStatement", "NotStatement", field(default=None)),
                            (
                                "ManagedRuleGroupStatement",
                                "ManagedRuleGroupStatement",
                                field(default=None),
                            ),
                            (
                                "LabelMatchStatement",
                                "LabelMatchStatement",
                                field(default=None),
                            ),
                            (
                                "RegexMatchStatement",
                                "RegexMatchStatement",
                                field(default=None),
                            ),
                        ],
                    ),
                ),
                (
                    "Action",
                    make_dataclass(
                        "RuleAction",
                        [
                            (
                                "Block",
                                make_dataclass(
                                    "BlockAction",
                                    [
                                        (
                                            "CustomResponse",
                                            make_dataclass(
                                                "CustomResponse",
                                                [
                                                    ("ResponseCode", int),
                                                    (
                                                        "CustomResponseBodyKey",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    (
                                                        "ResponseHeaders",
                                                        List[
                                                            make_dataclass(
                                                                "CustomHTTPHeader",
                                                                [
                                                                    ("Name", str),
                                                                    ("Value", str),
                                                                ],
                                                            )
                                                        ],
                                                        field(default=None),
                                                    ),
                                                ],
                                            ),
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "Allow",
                                make_dataclass(
                                    "AllowAction",
                                    [
                                        (
                                            "CustomRequestHandling",
                                            make_dataclass(
                                                "CustomRequestHandling",
                                                [
                                                    (
                                                        "InsertHeaders",
                                                        List["CustomHTTPHeader"],
                                                    )
                                                ],
                                            ),
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "Count",
                                make_dataclass(
                                    "CountAction",
                                    [
                                        (
                                            "CustomRequestHandling",
                                            "CustomRequestHandling",
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                            (
                                "Captcha",
                                make_dataclass(
                                    "CaptchaAction",
                                    [
                                        (
                                            "CustomRequestHandling",
                                            "CustomRequestHandling",
                                            field(default=None),
                                        )
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "OverrideAction",
                    make_dataclass(
                        "OverrideAction",
                        [
                            ("Count", "CountAction", field(default=None)),
                            ("none", Dict, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "VisibilityConfig",
                    make_dataclass(
                        "VisibilityConfig",
                        [
                            ("SampledRequestsEnabled", bool),
                            ("CloudWatchMetricsEnabled", bool),
                            ("MetricName", str),
                        ],
                    ),
                    field(default=None),
                ),
                (
                    "CaptchaConfig",
                    make_dataclass(
                        "CaptchaConfig",
                        [
                            (
                                "ImmunityTimeProperty",
                                make_dataclass(
                                    "ImmunityTimeProperty", [("ImmunityTime", int)]
                                ),
                                field(default=None),
                            )
                        ],
                    ),
                    field(default=None),
                ),
            ],
        )
    ] = None,
    custom_response_bodies: dict = None,
    captcha_config: dict = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
):
    r"""

    A web ACL defines a collection of rules to use to inspect and control web requests. Each rule has an action defined
    (allow, block, or count) for requests that match the statement of the rule. In the web ACL, you assign a default
    action to take (allow, block) for any request that does not match any of the rules. The rules in a web ACL can be
    a combination of the types Rule , RuleGroup , and managed rule group. You can associate a web ACL with one or more
    Amazon Web Services resources to protect. The resources can be an Amazon CloudFront distribution, an Amazon API
    Gateway REST API, an Application Load Balancer, or an AppSync GraphQL API.

    Args:
        name(text): The name of the web ACL. You cannot change the name of a web ACL after you create it.
        scope(text): Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
            A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API,
            or an AppSync GraphQL API.
            To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
              * CLI -- Specify the Region when you use the CloudFront scope: --scope=CLOUDFRONT --region=us-east-1 .
              * API and SDKs -- For all calls, use the Region endpoint us-east-1.
        default_action(Dict[str, Any]): The action to perform if none of the Rules contained in the WebACL match.
            * Block (Dict[str, Any], optional): Specifies that WAF should block requests by default.
                * CustomResponse (Dict[str, Any], optional): Defines a custom response for the web request. For information about customizing web requests
                    and responses, see Customizing web requests and responses in WAF in the WAF Developer Guide.
                    * ResponseCode (int): The HTTP status code to return to the client.  For a list of status codes that you can use in
                        your custom reqponses, see Supported status codes for custom response in the WAF Developer
                        Guide.
                    * CustomResponseBodyKey (str, optional): References the response body that you want WAF to return to the web request client. You can
                        define a custom response for a rule action or a default web ACL action that is set to block. To
                        do this, you first define the response body key and value in the CustomResponseBodies setting
                        for the WebACL or RuleGroup where you want to use it. Then, in the rule action or web ACL
                        default action BlockAction setting, you reference the response body using this key.
                    * ResponseHeaders (List[Dict[str, Any]], optional): The HTTP headers to use in the response. Duplicate header names are not allowed.  For
                        information about the limits on count and size for custom request and response settings, see WAF
                        quotas in the WAF Developer Guide.
                        * Name (str): The name of the custom header.  For custom request header insertion, when WAF inserts the header
                            into the request, it prefixes this name x-amzn-waf-, to avoid confusion with the headers that
                            are already in the request. For example, for the header name sample, WAF inserts the header
                            x-amzn-waf-sample.
                        * Value (str): The value of the custom header.
            * Allow (Dict[str, Any], optional): Specifies that WAF should allow requests by default.
                * CustomRequestHandling (Dict[str, Any], optional): Defines custom handling for the web request. For information about customizing web requests and
                    responses, see Customizing web requests and responses in WAF in the WAF Developer Guide.
                    * InsertHeaders (List['CustomHTTPHeader']): The HTTP headers to insert into the request. Duplicate header names are not allowed.  For
                        information about the limits on count and size for custom request and response settings, see WAF
                        quotas in the WAF Developer Guide.
        visibility_config(Dict[str, Any]): Defines and enables Amazon CloudWatch metrics and web request sample collection.
            * SampledRequestsEnabled (bool): A boolean indicating whether WAF should store a sampling of the web requests that match the
                rules. You can view the sampled requests through the WAF console.
            * CloudWatchMetricsEnabled (bool): A boolean indicating whether the associated resource sends metrics to Amazon CloudWatch. For the
                list of available metrics, see WAF Metrics.
            * MetricName (str): A name of the Amazon CloudWatch metric. The name can contain only the characters: A-Z, a-z, 0-9,
                - (hyphen), and _ (underscore). The name can be from one to 128 characters long. It can't
                contain whitespace or metric names reserved for WAF, for example "All" and "Default_Action."
        resource_id(text, optional): AWS WAF ID.
        description(text, optional): A description of the web ACL that helps with identification.
        rules(List[Dict[str, Any]], optional): The Rule statements used to identify the web requests that you want to allow, block, or count.
                Each rule includes one top-level statement that WAF uses to identify matching web requests, and
                parameters that govern how WAF handles them. Defaults to None.
            * Name (str): The name of the rule. You can't change the name of a Rule after you create it.
            * Priority (int): If you define more than one Rule in a WebACL, WAF evaluates each request against the Rules in
                order based on the value of Priority. WAF processes rules with lower priority first. The
                priorities don't need to be consecutive, but they must all be different.
                * Statement (Dict[str, Any]): The WAF processing statement for the rule, for example ByteMatchStatement or
                SizeConstraintStatement.
                * ByteMatchStatement (Dict[str, Any], optional): A rule statement that defines a string match search for WAF to apply to web requests. The byte
                    match statement provides the bytes to search for, the location in requests that you want WAF to
                    search, and other settings. The bytes to search for are typically a string that corresponds with
                    ASCII characters. In the WAF console and the developer guide, this is refered to as a string
                    match statement.
                    * SearchString (ByteString): A string value that you want WAF to search for.
                    * FieldToMatch (Dict[str, Any]): The part of the web request that you want WAF to inspect. For more information, see
                        FieldToMatch.
                        * SingleHeader (Dict[str, Any], optional): Inspect a single header. Provide the name of the header to inspect, for example, User-Agent or
                            Referer. This setting isn't case sensitive. Example JSON: "SingleHeader": { "Name": "haystack" }
                            Alternately, you can filter and inspect all headers with the Headers FieldToMatch setting.
                            * Name (str): The name of the query header to inspect.
                        * SingleQueryArgument (Dict[str, Any], optional): Inspect a single query argument. Provide the name of the query argument to inspect, such as
                            UserName or SalesRegion. The name can be up to 30 characters long and isn't case sensitive.
                            Example JSON: "SingleQueryArgument": { "Name": "myArgument" }
                            * Name (str): The name of the query argument to inspect.
                        * AllQueryArguments (Dict, optional): Inspect all query arguments.
                        * UriPath (Dict, optional): Inspect the request URI path. This is the part of the web request that identifies a resource,
                            for example, /images/daily-ad.jpg.
                        * QueryString (Dict, optional): Inspect the query string. This is the part of a URL that appears after a ? character, if any.
                        * Body (Dict[str, Any], optional): Inspect the request body as plain text. The request body immediately follows the request
                            headers. This is the part of a request that contains any additional data that you want to send
                            to your web server as the HTTP request body, such as data from a form.  Only the first 8 KB
                            (8192 bytes) of the request body are forwarded to WAF for inspection by the underlying host
                            service. For information about how to handle oversized request bodies, see the Body object
                            configuration.
                            * OversizeHandling (str, optional): What WAF should do if the body is larger than WAF can inspect.
                        * Method (Dict, optional): Inspect the HTTP method. The method indicates the type of operation that the request is asking
                            the origin to perform.
                        * JsonBody (Dict[str, Any], optional): Inspect the request body as JSON. The request body immediately follows the request headers. This
                            is the part of a request that contains any additional data that you want to send to your web
                            server as the HTTP request body, such as data from a form.  Only the first 8 KB (8192 bytes) of
                            the request body are forwarded to WAF for inspection by the underlying host service. For
                            information about how to handle oversized request bodies, see the JsonBody object configuration.
                            * MatchPattern (Dict[str, Any]): The patterns to look for in the JSON body. WAF inspects the results of these pattern matches
                                against the rule inspection criteria.
                                * All (Dict, optional): Match all of the elements. See also MatchScope in JsonBody.  You must specify either this
                                    setting or the IncludedPaths setting, but not both.
                                * IncludedPaths (List[str], optional): Match only the specified include paths. See also MatchScope in JsonBody.  Provide the include
                                    paths using JSON Pointer syntax. For example, "IncludedPaths": ["/dogs/0/name", "/dogs/1/name"].
                                    For information about this syntax, see the Internet Engineering Task Force (IETF) documentation
                                    JavaScript Object Notation (JSON) Pointer.  You must specify either this setting or the All
                                    setting, but not both.  Don't use this option to include all paths. Instead, use the All
                                    setting.
                            * MatchScope (str): The parts of the JSON to match against using the MatchPattern. If you specify All, WAF matches
                                against keys and values.
                            * InvalidFallbackBehavior (str, optional): What WAF should do if it fails to completely parse the JSON body.
                                The options are the following:
                                    EVALUATE_AS_STRING - Inspect the body as plain text. WAF applies the text transformations and
                                        inspection criteria that you defined for the JSON inspection to the body text string.
                                    MATCH - Treat the web request as matching the rule statement. WAF applies the rule action to the request.
                                    NO_MATCH - Treat the web request as not matching the rule statement.
                            * OversizeHandling (str, optional): What WAF should do if the body is larger than WAF can inspect.
                        * Headers (Dict[str, Any], optional): Inspect the request headers. You must configure scope and pattern matching filters in the
                            Headers object, to define the set of headers to and the parts of the headers that WAF inspects.
                            Only the first 8 KB (8192 bytes) of a request's headers and only the first 200 headers are
                            forwarded to WAF for inspection by the underlying host service. You must configure how to handle
                            any oversize header content in the Headers object. WAF applies the pattern matching filters to
                            the headers that it receives from the underlying host service.
                            * MatchPattern (Dict[str, Any]): The filter to use to identify the subset of headers to inspect in a web request.  You must
                                specify exactly one setting: either All, IncludedHeaders, or ExcludedHeaders. Example JSON:
                                "HeaderMatchPattern": { "ExcludedHeaders": {"KeyToExclude1", "KeyToExclude2"} }
                                * All (Dict, optional): Inspect all headers.
                                * IncludedHeaders (List[str], optional): Inspect only the headers that have a key that matches one of the strings specified here.
                                * ExcludedHeaders (List[str], optional): Inspect only the headers whose keys don't match any of the strings specified here.
                            * MatchScope (str): The parts of the headers to match with the rule inspection criteria. If you specify All, WAF
                                inspects both keys and values.
                            * OversizeHandling (str): What WAF should do if the headers of the request are larger than WAF can inspect.
                        * Cookies (Dict[str, Any], optional): Inspect the request cookies. You must configure scope and pattern matching filters in the
                            Cookies object, to define the set of cookies and the parts of the cookies that WAF inspects.
                            Only the first 8 KB (8192 bytes) of a request's cookies and only the first 200 cookies are
                            forwarded to WAF for inspection by the underlying host service. You must configure how to handle
                            any oversize cookie content in the Cookies object. WAF applies the pattern matching filters to
                            the cookies that it receives from the underlying host service.
                            * MatchPattern (Dict[str, Any]): The filter to use to identify the subset of cookies to inspect in a web request.  You must
                                specify exactly one setting: either All, IncludedCookies, or ExcludedCookies. Example JSON:
                                "CookieMatchPattern": { "IncludedCookies": {"KeyToInclude1", "KeyToInclude2", "KeyToInclude3"} }
                                * All (Dict, optional): Inspect all cookies.
                                * IncludedCookies (List[str], optional): Inspect only the cookies that have a key that matches one of the strings specified here.
                                * ExcludedCookies (List[str], optional): Inspect only the cookies whose keys don't match any of the strings specified here.
                            * MatchScope (str): The parts of the cookies to inspect with the rule inspection criteria. If you specify All, WAF
                                inspects both keys and values.
                            * OversizeHandling (str): What WAF should do if the cookies of the request are larger than WAF can inspect.
                    * TextTransformations (List[Dict[str, Any]]): Text transformations eliminate some of the unusual formatting that attackers use in web requests
                        in an effort to bypass detection. If you specify one or more transformations in a rule
                        statement, WAF performs all transformations on the content of the request component identified
                        by FieldToMatch, starting from the lowest priority setting, before inspecting the content for a
                        match.
                        * Priority (int): Sets the relative processing order for multiple transformations that are defined for a rule
                            statement. WAF processes all transformations, from lowest priority to highest, before inspecting
                            the transformed content. The priorities don't need to be consecutive, but they must all be
                            different.
                        * Type (str): You can specify the following transformation types:
                            BASE64_DECODE - Decode a Base64-encoded string.
                            BASE64_DECODE_EXT - Decode a Base64-encoded string, but use a forgiving implementation
                                that ignores characters that aren't valid.
                            CMD_LINE - Command-line transformations. These are helpful in reducing effectiveness of attackers who inject an operating system command-line
                                command and use unusual formatting to disguise some or all of the command.
                    * PositionalConstraint (str): The area within the portion of the web request that you want WAF to search for SearchString.
                        Valid values include the following:  CONTAINS  The specified part of the web request must
                        include the value of SearchString, but the location doesn't matter.  CONTAINS_WORD  The
                        specified part of the web request must include the value of SearchString, and SearchString must
                        contain only alphanumeric characters or underscore (A-Z, a-z, 0-9, or _). In addition,
                        SearchString must be a word, which means that both of the following are true:    SearchString is
                        at the beginning of the specified part of the web request or is preceded by a character other
                        than an alphanumeric character or underscore (_). Examples include the value of a header and
                        ;BadBot.    SearchString is at the end of the specified part of the web request or is followed
                        by a character other than an alphanumeric character or underscore (_), for example, BadBot; and
                        -BadBot;.    EXACTLY  The value of the specified part of the web request must exactly match the
                        value of SearchString.  STARTS_WITH  The value of SearchString must appear at the beginning of
                        the specified part of the web request.  ENDS_WITH  The value of SearchString must appear at the
                        end of the specified part of the web request.
                * SqliMatchStatement (Dict[str, Any], optional): Attackers sometimes insert malicious SQL code into web requests in an effort to extract data
                    from your database. To allow or block web requests that appear to contain malicious SQL code,
                    create one or more SQL injection match conditions. An SQL injection match condition identifies
                    the part of web requests, such as the URI or the query string, that you want WAF to inspect.
                    Later in the process, when you create a web ACL, you specify whether to allow or block requests
                    that appear to contain malicious SQL code.
                    * FieldToMatch ('FieldToMatch'): The part of the web request that you want WAF to inspect. For more information, see
                        FieldToMatch.
                    * TextTransformations (List['TextTransformation']): Text transformations eliminate some of the unusual formatting that attackers use in web requests
                        in an effort to bypass detection. If you specify one or more transformations in a rule
                        statement, WAF performs all transformations on the content of the request component identified
                        by FieldToMatch, starting from the lowest priority setting, before inspecting the content for a
                        match.
                * XssMatchStatement (Dict[str, Any], optional): A rule statement that defines a cross-site scripting (XSS) match search for WAF to apply to web
                    requests. XSS attacks are those where the attacker uses vulnerabilities in a benign website as a
                    vehicle to inject malicious client-site scripts into other legitimate web browsers. The XSS
                    match statement provides the location in requests that you want WAF to search and text
                    transformations to use on the search area before WAF searches for character sequences that are
                    likely to be malicious strings.
                    * FieldToMatch ('FieldToMatch'): The part of the web request that you want WAF to inspect. For more information, see
                        FieldToMatch.
                    * TextTransformations (List['TextTransformation']): Text transformations eliminate some of the unusual formatting that attackers use in web requests
                        in an effort to bypass detection. If you specify one or more transformations in a rule
                        statement, WAF performs all transformations on the content of the request component identified
                        by FieldToMatch, starting from the lowest priority setting, before inspecting the content for a
                        match.
                * SizeConstraintStatement (Dict[str, Any], optional): A rule statement that compares a number of bytes against the size of a request component, using
                        a comparison operator, such as greater than (>) or less than (<). For example, you can use a
                        size constraint statement to look for query strings that are longer than 100 bytes.  If you
                        configure WAF to inspect the request body, WAF inspects only the first 8192 bytes (8 KB). If the
                        request body for your web requests never exceeds 8192 bytes, you can create a size constraint
                        condition and block requests that have a request body greater than 8192 bytes. If you choose URI
                        for the value of Part of the request to filter on, the slash (/) in the URI counts as one
                        character. For example, the URI /logo.jpg is nine characters long.
                    * FieldToMatch ('FieldToMatch'): The part of the web request that you want WAF to inspect. For more information, see
                        FieldToMatch.
                    * ComparisonOperator (str): The operator to use to compare the request part to the size setting.
                    * Size (int): The size, in byte, to compare to the request part, after any transformations.
                    * TextTransformations (List['TextTransformation']): Text transformations eliminate some of the unusual formatting that attackers use in web requests
                        in an effort to bypass detection. If you specify one or more transformations in a rule
                        statement, WAF performs all transformations on the content of the request component identified
                        by FieldToMatch, starting from the lowest priority setting, before inspecting the content for a
                        match.
                * GeoMatchStatement (Dict[str, Any], optional): A rule statement used to identify web requests based on country of origin.
                    * CountryCodes (List[str], optional): An array of two-character country codes, for example, [ "US", "CN" ], from the alpha-2 country
                        ISO codes of the ISO 3166 international standard.
                    * ForwardedIPConfig (Dict[str, Any], optional): The configuration for inspecting IP addresses in an HTTP header that you specify, instead of
                        using the IP address that's reported by the web request origin. Commonly, this is the
                        X-Forwarded-For (XFF) header, but you can specify any header name.   If the specified header
                        isn't present in the request, WAF doesn't apply the rule to the web request at all.
                        * HeaderName (str): The name of the HTTP header to use for the IP address. For example, to use the X-Forwarded-For
                            (XFF) header, set this to X-Forwarded-For.  If the specified header isn't present in the
                            request, WAF doesn't apply the rule to the web request at all.
                        * FallbackBehavior (str): The match status to assign to the web request if the request doesn't have a valid IP address in
                            the specified position.  If the specified header isn't present in the request, WAF doesn't apply
                            the rule to the web request at all.  You can specify the following fallback behaviors:
                                MATCH - Treat the web request as matching the rule statement. WAF applies the rule action to the request.
                                NO_MATCH - Treat the web request as not matching the rule statement.
                * RuleGroupReferenceStatement (Dict[str, Any], optional): A rule statement used to run the rules that are defined in a RuleGroup. To use this, create a
                    rule group with your rules, then provide the ARN of the rule group in this statement. You cannot
                    nest a RuleGroupReferenceStatement, for example for use inside a NotStatement or OrStatement.
                    You can only use a rule group reference statement at the top level inside a web ACL.
                    * ARN (str): The Amazon Resource Name (ARN) of the entity.
                    * ExcludedRules (List[Dict[str, Any]], optional): The rules in the referenced rule group whose actions are set to Count. When you exclude a rule,
                        WAF evaluates it exactly as it would if the rule action setting were Count. This is a useful
                        option for testing the rules in a rule group without modifying how they handle your web traffic.
                        * Name (str): The name of the rule whose action you want to override to Count.
                * IPSetReferenceStatement (Dict[str, Any], optional): A rule statement used to detect web requests coming from particular IP addresses or address
                    ranges. To use this, create an IPSet that specifies the addresses you want to detect, then use
                    the ARN of that set in this statement. To create an IP set, see CreateIPSet. Each IP set rule
                    statement references an IP set. You create and maintain the set independent of your rules. This
                    allows you to use the single set in multiple rules. When you update the referenced set, WAF
                    automatically updates all rules that reference it.
                    * ARN (str): The Amazon Resource Name (ARN) of the IPSet that this statement references.
                    * IPSetForwardedIPConfig (Dict[str, Any], optional): The configuration for inspecting IP addresses in an HTTP header that you specify, instead of
                        using the IP address that's reported by the web request origin. Commonly, this is the
                        X-Forwarded-For (XFF) header, but you can specify any header name.   If the specified header
                        isn't present in the request, WAF doesn't apply the rule to the web request at all.
                        * HeaderName (str): The name of the HTTP header to use for the IP address. For example, to use the X-Forwarded-For
                            (XFF) header, set this to X-Forwarded-For.  If the specified header isn't present in the
                            request, WAF doesn't apply the rule to the web request at all.
                        * FallbackBehavior (str): The match status to assign to the web request if the request doesn't have a valid IP address in
                            the specified position.  If the specified header isn't present in the request, WAF doesn't apply
                            the rule to the web request at all.  You can specify the following fallback behaviors:
                                MATCH - Treat the web request as matching the rule statement. WAF applies the rule action to the request.
                                NO_MATCH - Treat the web request as not matching the rule statement.
                        * Position (str): The position in the header to search for the IP address. The header can contain IP addresses of
                            the original client and also of proxies. For example, the header value could be 10.1.1.1,
                            127.0.0.0, 10.10.10.10 where the first IP address identifies the original client and the rest
                            identify proxies that the request went through.  The options for this setting are the following:
                            FIRST - Inspect the first IP address in the list of IP addresses in the header. This is usually
                            the client's original IP.   LAST - Inspect the last IP address in the list of IP addresses in
                            the header.   ANY - Inspect all IP addresses in the header for a match. If the header contains
                            more than 10 IP addresses, WAF inspects the last 10.
                * RegexPatternSetReferenceStatement (Dict[str, Any], optional): A rule statement used to search web request components for matches with regular expressions. To
                    use this, create a RegexPatternSet that specifies the expressions that you want to detect, then
                    use the ARN of that set in this statement. A web request matches the pattern set rule statement
                    if the request component matches any of the patterns in the set. To create a regex pattern set,
                    see CreateRegexPatternSet. Each regex pattern set rule statement references a regex pattern set.
                    You create and maintain the set independent of your rules. This allows you to use the single set
                    in multiple rules. When you update the referenced set, WAF automatically updates all rules that
                    reference it.
                    * ARN (str): The Amazon Resource Name (ARN) of the RegexPatternSet that this statement references.
                    * FieldToMatch ('FieldToMatch'): The part of the web request that you want WAF to inspect. For more information, see
                        FieldToMatch.
                    * TextTransformations (List['TextTransformation']): Text transformations eliminate some of the unusual formatting that attackers use in web requests
                        in an effort to bypass detection. If you specify one or more transformations in a rule
                        statement, WAF performs all transformations on the content of the request component identified
                        by FieldToMatch, starting from the lowest priority setting, before inspecting the content for a
                        match.
                * RateBasedStatement (Dict[str, Any], optional): A rate-based rule tracks the rate of requests for each originating IP address, and triggers the
                    rule action when the rate exceeds a limit that you specify on the number of requests in any
                    5-minute time span. You can use this to put a temporary block on requests from an IP address
                    that is sending excessive requests.  WAF tracks and manages web requests separately for each
                    instance of a rate-based rule that you use. For example, if you provide the same rate-based rule
                    settings in two web ACLs, each of the two rule statements represents a separate instance of the
                    rate-based rule and gets its own tracking and management by WAF. If you define a rate-based rule
                    inside a rule group, and then use that rule group in multiple places, each use creates a
                    separate instance of the rate-based rule that gets its own tracking and management by WAF.  When
                    the rule action triggers, WAF blocks additional requests from the IP address until the request
                    rate falls below the limit. You can optionally nest another statement inside the rate-based
                    statement, to narrow the scope of the rule so that it only counts requests that match the nested
                    statement. For example, based on recent requests that you have seen from an attacker, you might
                    create a rate-based rule with a nested AND rule statement that contains the following nested
                    statements:   An IP match statement with an IP set that specified the address 192.0.2.44.   A
                    string match statement that searches in the User-Agent header for the string BadBot.   In this
                    rate-based rule, you also define a rate limit. For this example, the rate limit is 1,000.
                    Requests that meet both of the conditions in the statements are counted. If the count exceeds
                    1,000 requests per five minutes, the rule action triggers. Requests that do not meet both
                    conditions are not counted towards the rate limit and are not affected by this rule. You cannot
                    nest a RateBasedStatement inside another statement, for example inside a NotStatement or
                    OrStatement. You can define a RateBasedStatement inside a web ACL and inside a rule group.
                    * Limit (int): The limit on requests per 5-minute period for a single originating IP address. If the statement
                        includes a ScopeDownStatement, this limit is applied only to the requests that match the
                        statement.
                    * AggregateKeyType (str): Setting that indicates how to aggregate the request counts. The options are the following:
                       IP - Aggregate the request counts on the IP address from the web request origin.
                       FORWARDED_IP - Aggregate the request counts on the first IP address in an HTTP header. If you use this,
                        configure the ForwardedIPConfig, to specify the header to use.
                    * ScopeDownStatement ('Statement', optional): An optional nested statement that narrows the scope of the web requests that are evaluated by
                        the rate-based statement. Requests are only tracked by the rate-based statement if they match
                        the scope-down statement. You can use any nestable Statement in the scope-down statement, and
                        you can nest statements at any level, the same as you can for a rule statement.
                    * ForwardedIPConfig ('ForwardedIPConfig', optional): The configuration for inspecting IP addresses in an HTTP header that you specify, instead of
                        using the IP address that's reported by the web request origin. Commonly, this is the
                        X-Forwarded-For (XFF) header, but you can specify any header name.   If the specified header
                        isn't present in the request, WAF doesn't apply the rule to the web request at all.  This is
                        required if AggregateKeyType is set to FORWARDED_IP.
                * AndStatement (Dict[str, Any], optional): A logical rule statement used to combine other rule statements with AND logic. You provide more
                    than one Statement within the AndStatement.
                    * Statements (List[Dict[str, Any]]): The statements to combine with AND logic. You can use any statements that can be nested.
                        * ByteMatchStatement ('ByteMatchStatement', optional): A rule statement that defines a string match search for WAF to apply to web requests. The byte
                            match statement provides the bytes to search for, the location in requests that you want WAF to
                            search, and other settings. The bytes to search for are typically a string that corresponds with
                            ASCII characters. In the WAF console and the developer guide, this is refered to as a string
                            match statement.
                        * SqliMatchStatement ('SqliMatchStatement', optional): Attackers sometimes insert malicious SQL code into web requests in an effort to extract data
                            from your database. To allow or block web requests that appear to contain malicious SQL code,
                            create one or more SQL injection match conditions. An SQL injection match condition identifies
                            the part of web requests, such as the URI or the query string, that you want WAF to inspect.
                            Later in the process, when you create a web ACL, you specify whether to allow or block requests
                            that appear to contain malicious SQL code.
                        * XssMatchStatement ('XssMatchStatement', optional): A rule statement that defines a cross-site scripting (XSS) match search for WAF to apply to web
                            requests. XSS attacks are those where the attacker uses vulnerabilities in a benign website as a
                            vehicle to inject malicious client-site scripts into other legitimate web browsers. The XSS
                            match statement provides the location in requests that you want WAF to search and text
                            transformations to use on the search area before WAF searches for character sequences that are
                            likely to be malicious strings.
                        * SizeConstraintStatement ('SizeConstraintStatement', optional): A rule statement that compares a number of bytes against the size of a request component, using
                            a comparison operator, such as greater than (>) or less than (<). For example, you can use a
                            size constraint statement to look for query strings that are longer than 100 bytes.  If you
                            configure WAF to inspect the request body, WAF inspects only the first 8192 bytes (8 KB). If the
                            request body for your web requests never exceeds 8192 bytes, you can create a size constraint
                            condition and block requests that have a request body greater than 8192 bytes. If you choose URI
                            for the value of Part of the request to filter on, the slash (/) in the URI counts as one
                            character. For example, the URI /logo.jpg is nine characters long.
                        * GeoMatchStatement ('GeoMatchStatement', optional): A rule statement used to identify web requests based on country of origin.
                        * RuleGroupReferenceStatement ('RuleGroupReferenceStatement', optional): A rule statement used to run the rules that are defined in a RuleGroup. To use this, create a
                            rule group with your rules, then provide the ARN of the rule group in this statement. You cannot
                            nest a RuleGroupReferenceStatement, for example for use inside a NotStatement or OrStatement.
                            You can only use a rule group reference statement at the top level inside a web ACL.
                        * IPSetReferenceStatement ('IPSetReferenceStatement', optional): A rule statement used to detect web requests coming from particular IP addresses or address
                            ranges. To use this, create an IPSet that specifies the addresses you want to detect, then use
                            the ARN of that set in this statement. To create an IP set, see CreateIPSet. Each IP set rule
                            statement references an IP set. You create and maintain the set independent of your rules. This
                            allows you to use the single set in multiple rules. When you update the referenced set, WAF
                            automatically updates all rules that reference it.
                        * RegexPatternSetReferenceStatement ('RegexPatternSetReferenceStatement', optional): A rule statement used to search web request components for matches with regular expressions. To
                            use this, create a RegexPatternSet that specifies the expressions that you want to detect, then
                            use the ARN of that set in this statement. A web request matches the pattern set rule statement
                            if the request component matches any of the patterns in the set. To create a regex pattern set,
                            see CreateRegexPatternSet. Each regex pattern set rule statement references a regex pattern set.
                            You create and maintain the set independent of your rules. This allows you to use the single set
                            in multiple rules. When you update the referenced set, WAF automatically updates all rules that
                            reference it.
                        * RateBasedStatement ('RateBasedStatement', optional): A rate-based rule tracks the rate of requests for each originating IP address, and triggers the
                            rule action when the rate exceeds a limit that you specify on the number of requests in any
                            5-minute time span. You can use this to put a temporary block on requests from an IP address
                            that is sending excessive requests.  WAF tracks and manages web requests separately for each
                            instance of a rate-based rule that you use. For example, if you provide the same rate-based rule
                            settings in two web ACLs, each of the two rule statements represents a separate instance of the
                            rate-based rule and gets its own tracking and management by WAF. If you define a rate-based rule
                            inside a rule group, and then use that rule group in multiple places, each use creates a
                            separate instance of the rate-based rule that gets its own tracking and management by WAF.  When
                            the rule action triggers, WAF blocks additional requests from the IP address until the request
                            rate falls below the limit. You can optionally nest another statement inside the rate-based
                            statement, to narrow the scope of the rule so that it only counts requests that match the nested
                            statement. For example, based on recent requests that you have seen from an attacker, you might
                            create a rate-based rule with a nested AND rule statement that contains the following nested
                            statements:   An IP match statement with an IP set that specified the address 192.0.2.44.   A
                            string match statement that searches in the User-Agent header for the string BadBot.   In this
                            rate-based rule, you also define a rate limit. For this example, the rate limit is 1,000.
                            Requests that meet both of the conditions in the statements are counted. If the count exceeds
                            1,000 requests per five minutes, the rule action triggers. Requests that do not meet both
                            conditions are not counted towards the rate limit and are not affected by this rule. You cannot
                            nest a RateBasedStatement inside another statement, for example inside a NotStatement or
                            OrStatement. You can define a RateBasedStatement inside a web ACL and inside a rule group.
                        * AndStatement ('AndStatement', optional): A logical rule statement used to combine other rule statements with AND logic. You provide more
                            than one Statement within the AndStatement.
                        * OrStatement (Dict[str, Any], optional): A logical rule statement used to combine other rule statements with OR logic. You provide more
                            than one Statement within the OrStatement.
                            * Statements (List['Statement']): The statements to combine with OR logic. You can use any statements that can be nested.
                        * NotStatement (Dict[str, Any], optional): A logical rule statement used to negate the results of another rule statement. You provide one
                            Statement within the NotStatement.
                            * Statement ('Statement'): The statement to negate. You can use any statement that can be nested.
                        * ManagedRuleGroupStatement (Dict[str, Any], optional): A rule statement used to run the rules that are defined in a managed rule group. To use this,
                            provide the vendor name and the name of the rule group in this statement. You can retrieve the
                            required names by calling ListAvailableManagedRuleGroups. You cannot nest a
                            ManagedRuleGroupStatement, for example for use inside a NotStatement or OrStatement. It can only
                            be referenced as a top-level statement within a rule.
                            * VendorName (str): The name of the managed rule group vendor. You use this, along with the rule group name, to
                                identify the rule group.
                            * Name (str): The name of the managed rule group. You use this, along with the vendor name, to identify the
                                rule group.
                            * Version (str, optional): The version of the managed rule group to use. If you specify this, the version setting is fixed
                                until you change it. If you don't specify this, WAF uses the vendor's default version, and then
                                keeps the version at the vendor's default when the vendor updates the managed rule group
                                settings.
                            * ExcludedRules (List['ExcludedRule'], optional): The rules in the referenced rule group whose actions are set to Count. When you exclude a rule,
                                WAF evaluates it exactly as it would if the rule action setting were Count. This is a useful
                                option for testing the rules in a rule group without modifying how they handle your web traffic.
                            * ScopeDownStatement ('Statement', optional): An optional nested statement that narrows the scope of the web requests that are evaluated by
                                the managed rule group. Requests are only evaluated by the rule group if they match the scope-
                                down statement. You can use any nestable Statement in the scope-down statement, and you can nest
                                statements at any level, the same as you can for a rule statement.
                            * ManagedRuleGroupConfigs (List[Dict[str, Any]], optional): Additional information that's used by a managed rule group. Most managed rule groups don't
                                require this. Use this for the account takeover prevention managed rule group
                                AWSManagedRulesATPRuleSet, to provide information about the sign-in page of your application.
                                You can provide multiple individual ManagedRuleGroupConfig objects for any rule group
                                configuration, for example UsernameField and PasswordField. The configuration that you provide
                                depends on the needs of the managed rule group. For the ATP managed rule group, you provide the
                                following individual configuration objects: LoginPath, PasswordField, PayloadType and
                                UsernameField.
                                * LoginPath (str, optional): The path of the login endpoint for your application. For example, for the URL
                                    https://example.com/web/login, you would provide the path /web/login.
                                * PayloadType (str, optional): The payload type for your login endpoint, either JSON or form encoded.
                                * UsernameField (Dict[str, Any], optional): Details about your login page username field.
                                    * Identifier (str): The name of the username field. For example /form/username.
                                * PasswordField (Dict[str, Any], optional): Details about your login page password field.
                                    * Identifier (str): The name of the password field. For example /form/password.
                        * LabelMatchStatement (Dict[str, Any], optional): A rule statement that defines a string match search against labels that have been added to the
                            web request by rules that have already run in the web ACL.  The label match statement provides
                            the label or namespace string to search for. The label string can represent a part or all of the
                            fully qualified label name that had been added to the web request. Fully qualified labels have a
                            prefix, optional namespaces, and label name. The prefix identifies the rule group or web ACL
                            context of the rule that added the label. If you do not provide the fully qualified name in your
                            label match string, WAF performs the search for labels that were added in the same context as
                            the label match statement.
                            * Scope (str): Specify whether you want to match using the label name or just the namespace.
                            * Key (str): The string to match against. The setting you provide for this depends on the match statement's
                                Scope setting:    If the Scope indicates LABEL, then this specification must include the name
                                and can include any number of preceding namespace specifications and prefix up to providing the
                                fully qualified label name.    If the Scope indicates NAMESPACE, then this specification can
                                include any number of contiguous namespace strings, and can include the entire label namespace
                                prefix from the rule group or web ACL where the label originates.   Labels are case sensitive
                                and components of a label must be separated by colon, for example NS1:NS2:name.
                        * RegexMatchStatement (Dict[str, Any], optional): A rule statement used to search web request components for a match against a single regular
                            expression.
                            * RegexString (str): The string representing the regular expression.
                            * FieldToMatch ('FieldToMatch'): The part of the web request that you want WAF to inspect. For more information, see
                                FieldToMatch.
                            * TextTransformations (List['TextTransformation']): Text transformations eliminate some of the unusual formatting that attackers use in web requests
                                in an effort to bypass detection. If you specify one or more transformations in a rule
                                statement, WAF performs all transformations on the content of the request component identified
                                by FieldToMatch, starting from the lowest priority setting, before inspecting the content for a
                                match.
                * OrStatement ('OrStatement', optional): A logical rule statement used to combine other rule statements with OR logic. You provide more
                    than one Statement within the OrStatement.
                * NotStatement ('NotStatement', optional): A logical rule statement used to negate the results of another rule statement. You provide one
                    Statement within the NotStatement.
                * ManagedRuleGroupStatement ('ManagedRuleGroupStatement', optional): A rule statement used to run the rules that are defined in a managed rule group. To use this,
                    provide the vendor name and the name of the rule group in this statement. You can retrieve the
                    required names by calling ListAvailableManagedRuleGroups. You cannot nest a
                    ManagedRuleGroupStatement, for example for use inside a NotStatement or OrStatement. It can only
                    be referenced as a top-level statement within a rule.
                * LabelMatchStatement ('LabelMatchStatement', optional): A rule statement that defines a string match search against labels that have been added to the
                    web request by rules that have already run in the web ACL.  The label match statement provides
                    the label or namespace string to search for. The label string can represent a part or all of the
                    fully qualified label name that had been added to the web request. Fully qualified labels have a
                    prefix, optional namespaces, and label name. The prefix identifies the rule group or web ACL
                    context of the rule that added the label. If you do not provide the fully qualified name in your
                    label match string, WAF performs the search for labels that were added in the same context as
                    the label match statement.
                * RegexMatchStatement ('RegexMatchStatement', optional): A rule statement used to search web request components for a match against a single regular
                    expression.
            * Action (Dict[str, Any], optional): The action that WAF should take on a web request when it matches the rule statement. Settings at
                the web ACL level can override the rule action setting.  This is used only for rules whose
                statements do not reference a rule group. Rule statements that reference a rule group include
                RuleGroupReferenceStatement and ManagedRuleGroupStatement.  You must specify either this Action
                setting or the rule OverrideAction setting, but not both:   If the rule statement does not
                reference a rule group, use this rule action setting and not the rule override action setting.
                If the rule statement references a rule group, use the override action setting and not this
                action setting.
                * Block (Dict[str, Any], optional): Instructs WAF to block the web request.
                    * CustomResponse (Dict[str, Any], optional): Defines a custom response for the web request. For information about customizing web requests
                        and responses, see Customizing web requests and responses in WAF in the WAF Developer Guide.
                        * ResponseCode (int): The HTTP status code to return to the client.  For a list of status codes that you can use in
                            your custom reqponses, see Supported status codes for custom response in the WAF Developer
                            Guide.
                        * CustomResponseBodyKey (str, optional): References the response body that you want WAF to return to the web request client. You can
                            define a custom response for a rule action or a default web ACL action that is set to block. To
                            do this, you first define the response body key and value in the CustomResponseBodies setting
                            for the WebACL or RuleGroup where you want to use it. Then, in the rule action or web ACL
                            default action BlockAction setting, you reference the response body using this key.
                        * ResponseHeaders (List[Dict[str, Any]], optional): The HTTP headers to use in the response. Duplicate header names are not allowed.  For
                            information about the limits on count and size for custom request and response settings, see WAF
                            quotas in the WAF Developer Guide.
                            * Name (str): The name of the custom header.  For custom request header insertion, when WAF inserts the header
                                into the request, it prefixes this name x-amzn-waf-, to avoid confusion with the headers that
                                are already in the request. For example, for the header name sample, WAF inserts the header
                                x-amzn-waf-sample.
                            * Value (str): The value of the custom header.
                * Allow (Dict[str, Any], optional): Instructs WAF to allow the web request.
                    * CustomRequestHandling (Dict[str, Any], optional): Defines custom handling for the web request. For information about customizing web requests and
                        responses, see Customizing web requests and responses in WAF in the WAF Developer Guide.
                        * InsertHeaders (List['CustomHTTPHeader']): The HTTP headers to insert into the request. Duplicate header names are not allowed.  For
                            information about the limits on count and size for custom request and response settings, see WAF
                            quotas in the WAF Developer Guide.
                * Count (Dict[str, Any], optional): Instructs WAF to count the web request and allow it.
                    * CustomRequestHandling ('CustomRequestHandling', optional): Defines custom handling for the web request. For information about customizing web requests and
                        responses, see Customizing web requests and responses in WAF in the WAF Developer Guide.
                * Captcha (Dict[str, Any], optional): Instructs WAF to run a CAPTCHA check against the web request.
                    * CustomRequestHandling ('CustomRequestHandling', optional): Defines custom handling for the web request. For information about customizing web requests and
                        responses, see Customizing web requests and responses in WAF in the WAF Developer Guide.
            * OverrideAction (Dict[str, Any], optional): The action to use in the place of the action that results from the rule group evaluation. Set
                the override action to none to leave the result of the rule group alone. Set it to count to
                override the result to count only.  You can only use this for rule statements that reference a
                rule group, like RuleGroupReferenceStatement and ManagedRuleGroupStatement.   This option is
                usually set to none. It does not affect how the rules in the rule group are evaluated. If you
                want the rules in the rule group to only count matches, do not use this and instead exclude
                those rules in your rule group reference statement settings.
                * Count ('CountAction', optional): Override the rule group evaluation result to count only.   This option is usually set to none.
                    It does not affect how the rules in the rule group are evaluated. If you want the rules in the
                    rule group to only count matches, do not use this and instead exclude those rules in your rule
                    group reference statement settings.
                * None (Dict, optional): Don't override the rule group evaluation result. This is the most common setting.
            * RuleLabels (List[Dict[str, Any]], optional): Labels to apply to web requests that match the rule match statement. WAF applies fully qualified
                labels to matching web requests. A fully qualified label is the concatenation of a label
                namespace and a rule label. The rule's rule group or web ACL defines the label namespace.  Rules
                that run after this rule in the web ACL can match against these labels using a
                LabelMatchStatement. For each label, provide a case-sensitive string containing optional
                namespaces and a label name, according to the following guidelines:   Separate each component of
                the label with a colon.    Each namespace or name can have up to 128 characters.   You can
                specify up to 5 namespaces in a label.   Don't use the following reserved words in your label
                specification: aws, waf, managed, rulegroup, webacl, regexpatternset, or ipset.   For example,
                myLabelName or nameSpace1:nameSpace2:myLabelName.
                * Name (str): The label string.
            * VisibilityConfig (Dict[str, Any]): Defines and enables Amazon CloudWatch metrics and web request sample collection.
                * SampledRequestsEnabled (bool): A boolean indicating whether WAF should store a sampling of the web requests that match the
                    rules. You can view the sampled requests through the WAF console.
                * CloudWatchMetricsEnabled (bool): A boolean indicating whether the associated resource sends metrics to Amazon CloudWatch. For the
                    list of available metrics, see WAF Metrics.
                * MetricName (str): A name of the Amazon CloudWatch metric. The name can contain only the characters: A-Z, a-z, 0-9,
                    - (hyphen), and _ (underscore). The name can be from one to 128 characters long. It can't
                    contain whitespace or metric names reserved for WAF, for example "All" and "Default_Action."
            * CaptchaConfig (Dict[str, Any], optional): Specifies how WAF should handle CAPTCHA evaluations. If you don't specify this, WAF uses the
                CAPTCHA configuration that's defined for the web ACL.
                * ImmunityTimeProperty (Dict[str, Any], optional): Determines how long a CAPTCHA token remains valid after the client successfully solves a CAPTCHA
                    puzzle.
                    * ImmunityTime (int): The amount of time, in seconds, that a CAPTCHA token is valid. The default setting is 300.
        custom_response_bodies(dict, optional): A map of custom response keys and content bodies. When you create a rule with a
            block action, you can send a custom response to the web request. You define these for the web ACL, and then
            use them in the rules and default actions that you define in the web ACL.
        captcha_config(dict, optional): Specifies how WAF should handle CAPTCHA evaluations for rules that don't have their own
            CaptchaConfig settings. If you don't specify this, WAF uses its default settings for CaptchaConfig.
              * ImmunityTimeProperty (dict) -- Determines how long a CAPTCHA token remains valid after the client
                    successfully solves a CAPTCHA puzzle.
                  * ImmunityTime(integer) -- The amount of time, in seconds, that a CAPTCHA token is valid. The default
                        setting is 300.
        tags(list, optional): A tag associated with an Amazon Web Services resource. Tags are key:value pairs that you can use to
            categorize and manage your resources, for purposes like billing or other management. Typically, the tag key
            represents a category, such as "environment", and the tag value represents a specific value within that
            category, such as "test," "development," or "production". Or you might set the tag key to "customer" and
            the value to the customer name or ID. You can specify one or more tags to add to each Amazon Web Services
            resource, up to 50 tags for a resource.
            list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value}

    Request Syntax:
        [web-acl-name]::
            aws.wafv2.web_acl.present:
            - name: 'string'
            - scope: 'string'
            - default_action: 'dict'
            - description: 'string'
            - rules: ['string']
            - visibility_config: 'dict'
            - custom_response_bodies: 'dict'
            - captcha_config: 'dict'
            - tags:
              - Key: 'string'
                Value: 'string'

    Examples:

        .. code-block:: sls
        demo_waf:
            aws.wafv2.web_acl.present:
            - name: demo_waf.
            - scope: REGIONAL
            - default_action:
                Allow: {}
            - description: waf to support REGIONAL resources.
            - rules:
                - Name: AWS-AWSManagedRulesBotControlRuleSet
                  OverrideAction:
                  None: {}
                  Priority: 1
                  Statement:
                    ManagedRuleGroupStatement:
                    Name: AWSManagedRulesBotControlRuleSet
                    VendorName: AWS
                  VisibilityConfig:
                    CloudWatchMetricsEnabled: false
                    MetricName: AWS-AWSManagedRulesBotControlRuleSet
                    SampledRequestsEnabled: true
            - visibility_config:
                CloudWatchMetricsEnabled: false
                MetricName: test_metric_waf
                SampledRequestsEnabled: true
            - tags:
              - Key: type
                Value: REGIONAL

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    lock_token: str = None
    resource_updated: bool = False
    tags_list = None
    tags_dict = None
    if tags is not None:
        if isinstance(tags, Dict):
            tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            tags_dict = tags
        else:
            tags_list = tags
            tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    resource_parameters = {
        "Name": name,
        "DefaultAction": default_action,
        "VisibilityConfig": visibility_config,
        "Description": description,
        "Rules": rules,
        "CustomResponseBodies": custom_response_bodies,
        "CaptchaConfig": captcha_config,
        "Tags": tags_list,
    }

    if resource_id:
        ret = await hub.exec.boto3.client.wafv2.get_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id
        )
        if ret["result"]:
            before = ret["ret"]["WebACL"]
            lock_token = ret["ret"]["LockToken"]

    if before:
        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=before, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        # Update web acl
        update_ret = await hub.exec.aws.wafv2.web_acl.update_web_acl(
            ctx,
            name=name,
            raw_resource=before,
            resource_parameters=resource_parameters,
            scope=scope,
            resource_id=resource_id,
            lock_token=lock_token,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])

        if update_ret["ret"] and ctx.get("test", False):
            for key in [
                "scope",
                "default_action",
                "visibility_config",
                "description",
                "rules",
                "custom_response_bodies",
                "captcha_config",
            ]:
                if key in update_ret["ret"]:
                    plan_state[key] = update_ret["ret"][key]

        if (tags_list is not None) and (
            not hub.tool.aws.state_comparison_utils.are_lists_identical(
                tags_list,
                hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    result["old_state"].get("tags")
                ),
            )
        ):
            # Update tags
            update_tag_ret = await hub.exec.aws.wafv2.web_acl.update_web_acl_tags(
                ctx=ctx,
                web_acl_arn=before.get("ARN"),
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags_dict,
            )
            result["result"] = result["result"] and update_tag_ret["result"]
            result["comment"] = result["comment"] + update_tag_ret["comment"]
            resource_updated = resource_updated or bool(update_tag_ret["ret"])

            if ctx.get("test", False) and update_tag_ret["ret"] is not None:
                plan_state["tags"] = update_tag_ret["ret"]

        if not resource_updated:
            result["comment"] = result["comment"] + (
                f"aws.wafv2.web_acl.present '{name}' has no property to update.",
            )

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "scope": scope,
                    "default_action": default_action,
                    "visibility_config": visibility_config,
                    "description": description,
                    "rules": rules,
                    "custom_response_bodies": custom_response_bodies,
                    "captcha_config": captcha_config,
                    "tags": tags_dict,
                },
            )
            result["comment"] = (f"Would create aws.wafv2.web_acl {name}",)
            return result

        # Create web acl
        ret = await hub.exec.boto3.client.wafv2.create_web_acl(
            ctx, Scope=scope, **resource_parameters
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result
        result["comment"] = result["comment"] + (f"Created '{name}'",)
        resource_id = ret["ret"]["Summary"]["Id"]

    if ctx.get("test", False):
        result["new_state"] = plan_state

    elif (not before) or resource_updated:
        resource_ret = await hub.exec.boto3.client.wafv2.get_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id
        )
        if not result["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result
        after = resource_ret["ret"]["WebACL"]

        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=after, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["new_state"] = convert_ret["ret"]

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub, ctx, name: str, scope: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes the specified WebACL.
    You can only use this if ManagedByFirewallManager is false in the specified WebACL.

    Args:
        name(text): The name of the web ACL. You cannot change the name of a web ACL after you create it.
        resource_id(text): AWS WAF ID. Idem automatically considers this resource being absent if this field is not specified.
        scope(text): Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
            A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API, or an
            AppSync GraphQL API.
            To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
            * CLI -- Specify the Region when you use the CloudFront scope: --scope=CLOUDFRONT --region=us-east-1 .
            * API and SDKs -- For all calls, use the Region endpoint us-east-1.

    Request Syntax:
        [web-acl-name]::
            aws.wafv2.web_acl.absent:
            - name: 'string'
            - scope: 'string'

    Examples:

        .. code-block:: sls

            demo_waf::
               aws.wafv2.web_acl.absent:
                - name: demo_waf
                - scope: REGIONAL

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    lock_token: str = None

    if resource_id:
        resource = await hub.exec.boto3.client.wafv2.get_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id
        )
        if resource["result"]:
            before = resource["ret"]["WebACL"]
            lock_token = resource["ret"]["LockToken"]

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            "aws.wafv2.web_acl", name
        )

    elif ctx.get("test", False):
        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=before, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            "aws.wafv2.web_acl", name
        )
        return result

    else:
        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=before, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"]

        # Delete web acl
        ret = await hub.exec.boto3.client.wafv2.delete_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id, LockToken=lock_token
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            "aws.wafv2.web_acl", name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Retrieves an WebACLSummary objects for the web ACLs that you manage.

    Examples:

        .. code-block:: bash

            $ idem describe aws.wafv2.web_acl

    Returns:
        Dict[str, Any]

    """
    result = {}
    scope = ["CLOUDFRONT", "REGIONAL"]

    for web_acl_scope in scope:
        ret = await hub.exec.boto3.client.wafv2.list_web_acls(ctx, Scope=web_acl_scope)
        if not ret["result"]:
            hub.log.debug(f"Could not describe web acl {ret['comment']}")
            continue

        for resource in ret["ret"]["WebACLs"]:
            web_acl_name = resource["Name"]
            resource_id = resource["Id"]
            raw_resource = await hub.exec.boto3.client.wafv2.get_web_acl(
                ctx, Name=web_acl_name, Id=resource_id, Scope=web_acl_scope
            )
            if not raw_resource["result"]:
                hub.log.warning(
                    f"Could not get web acl '{web_acl_name}' with error {convert_ret['comment']}"
                )
                continue
            resource_ret = raw_resource["ret"]["WebACL"]
            convert_ret = await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx,
                raw_resource=resource_ret,
                idem_resource_name=web_acl_name,
                scope=web_acl_scope,
            )
            if not convert_ret["result"]:
                hub.log.warning(
                    f"Could not describe web acl '{web_acl_name}' with error {convert_ret['comment']}"
                )
                continue
            translated_resource = convert_ret["ret"]
            result[translated_resource["resource_id"]] = {
                "aws.wafv2.web_acl.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in translated_resource.items()
                ]
            }

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions to convert raw resource state from AWS ECR to present input format.
"""


def convert_raw_repository_to_present(
    hub, raw_resource: Dict[str, Any], tags: List = None, idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("repositoryName")
    resource_parameters = OrderedDict(
        {
            "repositoryArn": "repository_arn",
            "registryId": "registry_id",
            "repositoryName": "repository_name",
            "repositoryUri": "repository_uri",
            "imageTagMutability": "image_tag_mutability",
            "imageScanningConfiguration": "image_scanning_configuration",
            "encryptionConfiguration": "encryption_configuration",
        }
    )

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if tags:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            tags
        )

    return resource_translated


def convert_raw_repository_policy_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    r"""
    Convert AWS ECR Repository Policy response to a common idem present state

    Args:
        hub: required for functions in hub
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(string, optional): An Idem name of the resource.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_id = (
        f"{raw_resource.get('registryId')}-{raw_resource.get('repositoryName')}"
    )
    resource_parameters = OrderedDict(
        {
            "registryId": "registry_id",
            "repositoryName": "repository_name",
        }
    )
    resource_translated = {
        "name": idem_resource_name if idem_resource_name else resource_id,
        "resource_id": resource_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("policyText"):
        resource_translated[
            "policy_text"
        ] = hub.tool.aws.state_comparison_utils.standardise_json(
            raw_resource.get("policyText")
        )

    return resource_translated



def get(hub, ctx) -> str:
    """
    Get the region name from ctx and fall back to the region in config
    """
    opt = getattr(hub, "OPT") or {}
    acct = opt.get("acct") or {}
    extras = acct.get("extras") or {}
    aws_opts = extras.get("aws") or {}

    return ctx.acct.get("region_name") or aws_opts.get("region_name")



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.organizations.deregister_delegated_administrator
hub.exec.boto3.client.organizations.list_delegated_administrators
hub.exec.boto3.client.organizations.register_delegated_administrator
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
SERVICE = "organizations"


TREQ = {
    "absent": {
        "require": [
            "aws.organizations.policy_attachment.absent",
            "aws.organizations.policy.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.organization_unit.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    email: str,
    role_name: str,
    iam_user_access_to_billing: str = "ALLOW",
    resource_id: str = None,
    parent_id: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an AWS account that is automatically a member of the organization whose credentials made the request.
    This is an asynchronous request that AWS performs in the background. Because CreateAccount operates asynchronously,
    it can return a successful completion message even though account initialization might still be in progress.
     You might need to wait a few minutes before you can successfully access the account

    Args:
        name(Text): The friendly name of the member account.
        email(Text): The email address of the owner to assign to the new member account. This email
                              address must not already be associated with another AWS account
        role_name(Text,Optional): The name of an IAM role that AWS Organizations automatically preconfigures in the
                                new member account. This role trusts the management account, allowing users in the
                                management account to assume the role, as permitted by the management account
                                administrator. The role has administrator permissions in the new member account.
                                If you don't specify this parameter, the role name defaults to OrganizationAccountAccessRole.
        iam_user_access_to_billing(Text,Optional,Default:'ALLOW'): If set to ALLOW , the new account enables IAM users to access account
                                                              billing information if they have the required permissions. If set to DENY ,
                                                              only the root user of the new account can access account billing information.
        resource_id(Text,Optional): AWS account ID to identify the resource
        parent_id(Text,Optional): Parent Organizational Unit ID or Root ID for the account. Defaults to the Organization default Root ID
        tags(List or Dict, optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                                      {tag-key: tag-value}

    Request Syntax:
        [account-id]:
          aws.organizations.account.present:
          - name: 'string'
          - resource_id: 'string'
          - email: 'string'
          - role_name: 'string'
          - iam_user_access_to_billing: 'string'
          - parent_id: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            764144542382:
                aws.organizations.account.present:
                    - name: name_of_new_account
                    - resource_id: new-account_id
                    - email: xyz@email.com
                    - role_name: role_name1
                    - iam_user_access_to_billing: ALLOW
                    - parent_id : ou108811
                    - tags:
                        - Key: test-key
                          Value: test-value
                        - Key: test-key-1
                          Value: test-key-1
    """

    result = dict(comment=(), name=name, result=True, old_state=None, new_state=None)
    before = None
    update_tag = False
    update_parent = False

    if resource_id:
        before = await hub.exec.boto3.client.organizations.describe_account(
            ctx, AccountId=resource_id
        )

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before:
        # Account exists , update
        try:
            old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
                ctx, ResourceId=resource_id
            )
            result["result"] = result["result"] and old_tags["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + old_tags["comment"]
                return result
            old_tags = old_tags["ret"].get("Tags") if old_tags else None
            # we need to list parents to check if move is required in case the new_parent_id is not equal to current_
            # parent_id
            parents = await hub.exec.boto3.client.organizations.list_parents(
                ctx, ChildId=resource_id
            )
            result["result"] = result["result"] and parents["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + parents["comment"]
                return result
            if parents:
                current_parent_id = parents["ret"]["Parents"][0]["Id"]
                if parent_id is not None and current_parent_id != parent_id:
                    if not ctx.get("test", False):
                        move_account_result = await move_account(
                            hub, ctx, resource_id, result, parent_id
                        )
                        if move_account_result and not move_account_result["result"]:
                            result["comment"] = result["comment"] + (
                                f"Could not update parent for aws.organizations.account {name}",
                            )
                            result["result"] = False
                            return result
                    update_parent = True
            result[
                "old_state"
            ] = hub.tool.aws.organization.conversion_utils.convert_raw_account_to_present(
                current_parent_id if parents else None,
                before["ret"]["Account"],
                old_tags if old_tags else None,
            )
            plan_state = copy.deepcopy(result["old_state"])
            if update_parent:
                plan_state["parent_id"] = parent_id
                result["comment"] = result["comment"] + (
                    f"Would update parent for aws.organizations.account {name}",
                )

            if tags is not None and tags != result["old_state"].get("tags"):
                update_tags_ret = (
                    await hub.exec.aws.organizations.organization.update_tags(
                        ctx, resource_id, result["old_state"].get("tags"), tags
                    )
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = update_tags_ret["result"]
                    return result

                result["comment"] = result["comment"] + (
                    f"Updated tags on aws.organizations.account '{name}'.",
                )
                update_tag = True
                if ctx.get("test", False):
                    if update_tags_ret["ret"]:
                        plan_state["tags"] = update_tags_ret["ret"]
                    else:
                        plan_state["tags"] = result["old_state"]["tags"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    else:
        # Account not present , create
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "email": email,
                    "name": name,
                    "parent_id": parent_id,
                    "role_name": role_name,
                    "iam_user_access_to_billing": iam_user_access_to_billing,
                    "tags": tags,
                },
            )
            result["comment"] = (f"Would create aws.organizations.account {name}",)
            return result

        try:
            create_account_ret = (
                await hub.exec.boto3.client.organizations.create_account(
                    ctx,
                    Email=email,
                    AccountName=name,
                    RoleName=role_name,
                    IamUserAccessToBilling=iam_user_access_to_billing,
                    Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                    if tags
                    else None,
                )
            )

            result["result"] = create_account_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + create_account_ret["comment"]
                return result

            account_status_id = create_account_ret["ret"]["CreateAccountStatus"]["Id"]

            # Call a custom waiter to wait on account's creation.

            acceptors = [
                {
                    "matcher": "path",
                    "expected": "SUCCEEDED",
                    "state": "success",
                    "argument": "CreateAccountStatus.State",
                },
                {
                    "matcher": "path",
                    "expected": "IN_PROGRESS",
                    "state": "retry",
                    "argument": "CreateAccountStatus.State",
                },
                {
                    "matcher": "path",
                    "expected": "FAILED",
                    # Failure is also mapped with success to catch the error message
                    "state": "success",
                    "argument": "CreateAccountStatus.State",
                },
            ]
            account_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="AccountCreated",
                operation="DescribeCreateAccountStatus",
                argument=["CreateAccountStatus.State"],
                acceptors=acceptors,
                client=hub.tool.boto3.client.get_client(ctx, SERVICE),
                matcher="path",
                delay=10,
                max_tries=10,
            )
            await hub.tool.boto3.client.wait(
                ctx,
                SERVICE,
                "AccountCreated",
                account_waiter,
                CreateAccountRequestId=account_status_id,
            )

            account_status = await hub.exec.boto3.client.organizations.describe_create_account_status(
                ctx, CreateAccountRequestId=account_status_id
            )
            if account_status["result"]:
                create_account_status = account_status["ret"]["CreateAccountStatus"]
                account_state = create_account_status["State"]
                if account_state == "FAILED":
                    result["result"] = False
                    result["comment"] = result["comment"] + (
                        create_account_status["FailureReason"],
                    )
                    return result
                elif account_state == "SUCCEEDED":
                    resource_id = create_account_status["AccountId"]
                    result["comment"] = (f"Created aws.organizations.account {name}.",)
                    if resource_id is not None and parent_id is not None:
                        parents = (
                            await hub.exec.boto3.client.organizations.list_parents(
                                ctx, ChildId=resource_id
                            )
                        )
                        result["result"] = result["result"] and parents["result"]
                        if not result["result"]:
                            result["comment"] = result["comment"] + parents["comment"]
                            return result
                        if parents:
                            current_parent_id = parents["ret"]["Parents"][0]["Id"]

                            if current_parent_id != parent_id:
                                move_account_result = await move_account(
                                    hub, ctx, resource_id, result, parent_id
                                )
                                if (
                                    move_account_result
                                    and not move_account_result["result"]
                                ):
                                    result["comment"] = result["comment"] + (
                                        f"Could not update parent for aws.organizations.account {name}",
                                    )
                                    result["result"] = False
                                    return result
            else:
                result["result"] = False
                result["comment"] = result["comment"] + account_status["comment"]
                return result

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not before or update_parent or update_tag and resource_id:
        try:
            after = await hub.exec.boto3.client.organizations.describe_account(
                ctx, AccountId=resource_id
            )
            if after and after.get("ret"):

                parents = await hub.exec.boto3.client.organizations.list_parents(
                    ctx, ChildId=resource_id
                )
                result["result"] = result["result"] and parents["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + parents["comment"]
                    return result
                if parents:
                    current_parent_id = parents["ret"]["Parents"][0]["Id"]

                new_tags = (
                    await hub.exec.boto3.client.organizations.list_tags_for_resource(
                        ctx, ResourceId=resource_id
                    )
                )
                result["result"] = result["result"] and new_tags["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + new_tags["comment"]
                    return result
                result[
                    "new_state"
                ] = hub.tool.aws.organization.conversion_utils.convert_raw_account_to_present(
                    current_parent_id if parents else None,
                    after["ret"]["Account"],
                    new_tags["ret"].get("Tags") if new_tags else None,
                )
            else:
                result["result"] = result["result"] and after["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + after["comment"]
                    return result

        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Removes the specified account from the organization.The removed account becomes a standalone
    account that isn't a member of any organization. It's no longer subject to any policies and
    is responsible for its own bill payments. The organization's management account is no longer
    charged for any expenses accrued by the member account after it's removed from the organization.
    This operation can be called only from the organization's management account. Member accounts
    can remove themselves with LeaveOrganization instead.

    Args:
        name(Text): Name of the member account
        resource_id(Text): AWS account ID to identify the resource


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            [account_id]:
              aws.organizations.organization.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_account(
        ctx, AccountId=resource_id
    )

    if not before:
        result["comment"] = (f"aws.organizations.account {name} already absent",)
    elif ctx.get("test", False):
        result["comment"] = (f"Would delete aws.organizations.account {name}.",)
    else:

        try:
            ret = await hub.exec.boto3.client.organizations.remove_account_from_organization(
                ctx, AccountId=resource_id
            )

            result["result"] = ret["result"]

            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = result["comment"] + (
                f"aws.organizations.account {name} deleted.",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if before:
        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_account_to_present(
            None, before["ret"]["Account"], None
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Lists all the accounts in the organization. To request only the accounts in a specified root or organizational
    unit (OU), use the ListAccountsForParent operation instead.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.account
    """

    result = {}

    describe_ret = await hub.exec.boto3.client.organizations.list_accounts(ctx)
    if not describe_ret:
        hub.log.debug(f"Could not describe account {describe_ret['comment']}")
        return {}

    accounts = describe_ret["ret"]["Accounts"]

    for account in accounts:
        parent = await hub.exec.boto3.client.organizations.list_parents(
            ctx, ChildId=account["Id"]
        )
        if parent and not parent["result"]:
            hub.log.debug(
                f"Unable to list parent for account: {account['Id']} Error:  {describe_ret['comment']}"
            )
            continue

        tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=account["Id"]
        )
        if tags and not tags["result"]:
            hub.log.debug(
                f"Unable to list tags for account: {account['Id']} Error:  {tags['comment']}"
            )
            continue
        translated_resource = (
            hub.tool.aws.organization.conversion_utils.convert_raw_account_to_present(
                parent["ret"]["Parents"][0]["Id"]
                if parent and parent["ret"].get("Parents")
                else None,
                account,
                tags["ret"].get("Tags") if tags and tags["ret"].get("Tags") else None,
            )
        )

        result[account["Id"]] = {
            "aws.organizations.account.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result


async def move_account(
    hub, ctx, account_id, result: Dict[str, Any], destination_parent_id
) -> Dict[str, Any]:
    if destination_parent_id is not None:

        parents = await hub.exec.boto3.client.organizations.list_parents(
            ctx, ChildId=account_id
        )
        result["result"] = result["result"] and parents["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + parents["comment"]
            return result

        if parents:
            parent_id = parents["ret"]["Parents"][0]["Id"]

            if parent_id == destination_parent_id:
                result["comment"] = result["comment"] + (
                    f"aws.organizations.account {account_id} already at {destination_parent_id}",
                )
                return result
            else:

                account_move_ret = (
                    await hub.exec.boto3.client.organizations.move_account(
                        ctx,
                        AccountId=account_id,
                        SourceParentId=parent_id,
                        DestinationParentId=destination_parent_id,
                    )
                )
                if not account_move_ret:
                    result["comment"] = result["comment"] + (
                        f"Could not move aws.organizations.account {account_id} to {destination_parent_id} Error: {account_move_ret['comment']}",
                    )

                    result["result"] = False

                    return result
                result["comment"] = result["comment"] + (
                    f"Successfully moved aws.organizations.account {account_id} under {destination_parent_id}",
                )
                return result



"""
Contains functions that are useful for describing instances in a consistent manner
"""
import asyncio
from typing import Any
from typing import Dict
from typing import Tuple


async def convert_to_present(
    hub, ctx, describe_instances: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert instances from ec2.describe_instances() to aws.ec2.instance.present states

    This is the preferred "meta" function for collecting information about multiple instances
    """
    result = {}
    coros = []
    for reservation in describe_instances.get("Reservations", []):
        for instance in reservation["Instances"]:
            instance_id = instance["InstanceId"]
            result[instance_id] = dict(
                name=instance_id,
                resource_id=instance_id,
                image_id=instance.get("ImageId"),
                instance_type=instance.get("InstanceType"),
                block_device_mappings=instance.get("BlockDeviceMappings"),
                ebs_optimized=instance.get("EbsOptimized"),
                kernel_id=instance.get("KernelId"),
                placement=instance.get("Placement"),
                subnet_id=instance.get("SubnetId"),
                network_interfaces=instance.get("NetworkInterfaces"),
                private_ip_address=instance.get("PrivateIpAddress"),
                monitoring_enabled=instance.get("Monitoring", {}).get("State")
                == "enabled",
                root_device_name=instance.get("RootDeviceName"),
                product_codes=instance.get("ProductCodes"),
                source_dest_check=instance.get("SourceDestCheck"),
                running=instance.get("State", {}).get("Name") == "running",
                reservation_id=reservation.get("ReservationId"),
                owner_id=reservation.get("OwnerId"),
                bootstrap=[],
            )

            coros.append(
                # launch_template config get most parameters for an instance
                hub.tool.aws.ec2.instance.state.config(ctx, instance_id=instance_id)
            )

    # This can be a heavy process if there are many instances, use coroutines to collect them all at simultaneously
    for ret in asyncio.as_completed(coros):
        instance_id, config = await ret
        result[instance_id].update(config)

    return hub.tool.aws.ec2.instance.data.sanitize_dict(result)


async def config(hub, ctx, *, instance_id: str) -> Tuple[str, Dict[str, Any]]:
    """
    Retrieves the configuration data of the specified instance.

    This action calls on other describe actions to get instance information.
    Depending on your instance configuration, you may need to allow the following actions in your IAM policy:
        - DescribeSpotInstanceRequests
        - DescribeInstanceCreditSpecifications
        - DescribeVolumes
        - DescribeInstanceAttribute
        - DescribeElasticGpus.
    Or, you can allow describe* depending on your instance requirements.
    """
    config = {}
    response = await hub.exec.boto3.client.ec2.get_launch_template_data(
        ctx, InstanceId=instance_id
    )
    if response:
        hub.log.trace(
            f"Collecting instance '{instance_id}' config from launch template data"
        )
        # This is ideal and concise
        config.update(
            hub.tool.aws.ec2.instance.state.parse_launch_template_data(
                **response.ret["LaunchTemplateData"]
            )
        )
    else:
        # Maybe we lack permissions, or we are using localstack and get_launch_template_data is not implemented yet
        # Get that same information elsewhere
        hub.log.trace(f"Collecting instance '{instance_id}' config manually")
        config.update(
            await hub.tool.aws.ec2.instance.state.attributes(
                ctx, instance_id=instance_id
            )
        )

    # The following attributes are not part of the launch template data
    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="sriovNetSupport", InstanceId=instance_id
    )
    if ret:
        config["sriov_net_support"] = ret.ret["SriovNetSupport"].get("Value")

    # describing "enaSupport" is not supported by AWS yet
    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="enaSupport", InstanceId=instance_id
    )
    if ret:
        config["ena_support"] = ret.ret["EnaSupport"].get("Value")

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="ebsOptimized", InstanceId=instance_id
    )
    if ret:
        config["ebs_optimized"] = ret.ret["EbsOptimized"].get("Value")

    return instance_id, config


def parse_launch_template_data(hub, **launch_config) -> Dict[str, Any]:
    """
    Parse LaunchTemplateData to collect information about a single instance
    """
    return dict(
        image_id=launch_config.get("ImageId"),
        instance_type=launch_config.get("InstanceType"),
        user_data=launch_config.get("UserData"),
        disable_api_termination=launch_config.get("DisableApiTermination"),
        instance_initiated_shutdown_behavior=launch_config.get(
            "InstanceInitiatedShutdownBehavior"
        ),
        block_device_mappings=launch_config.get("BlockDeviceMappings"),
        ebs_optimized=launch_config.get("EbsOptimized"),
        enclave_options_enabled=launch_config.get("EnclaveOptions", {}).get("Enabled"),
        kernel_id=launch_config.get("KernelId"),
        ram_disk_id=launch_config.get("RamDiskId"),
        monitoring_enbaled=launch_config.get("Monitoring", {}).get("Enabled"),
        network_interfaces=launch_config.get("NetworkInterfaces"),
        tags={
            tag["Key"]: tag.get("Value")
            for tag in launch_config.get("TagSpecifications", {}).get("Tags", [])
        },
        placement=launch_config.get("Placement"),
        elastic_gpu_specifications=launch_config.get("ElasticGpuSpecifications"),
        elastic_inference_accelerators=launch_config.get(
            "ElasticInferenceAccelerators"
        ),
        security_group_ids=launch_config.get("SecurityGroupIds"),
        instance_market_options=launch_config.get("InstanceMarketOptions"),
        credit_specification=launch_config.get("CreditSpecification"),
        cpu_options=launch_config.get("CpuOptions"),
        capacity_reservation_specification=launch_config.get(
            "CapacityReservationSpecification"
        ),
        license_specifications=launch_config.get("LicenseSpecifications"),
        hibernation_options=launch_config.get("HibernationOptions"),
        metadata_options=launch_config.get("MetadataOptions"),
        instance_requirements=launch_config.get("InstanceRequirements"),
        private_dns_name_options=launch_config.get("PrivateDnsNameOptions"),
        maintentance_options=launch_config.get("MaintenanceOptions"),
    )


async def attributes(hub, ctx, *, instance_id: str) -> Dict[str, Any]:
    """
    Manually collect information about a single instance
    """
    instance = {}

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="userData", InstanceId=instance_id
    )
    if ret:
        instance["user_data"] = ret.ret["UserData"]

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="disableApiTermination", InstanceId=instance_id
    )
    if ret:
        instance["disable_api_termination"] = ret.ret["DisableApiTermination"].get(
            "Value"
        )

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="instanceInitiatedShutdownBehavior", InstanceId=instance_id
    )
    if ret:
        instance["instance_initiated_shutdown_behavior"] = ret.ret[
            "InstanceInitiatedShutdownBehavior"
        ].get("Value")

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="enclaveOptions", InstanceId=instance_id
    )
    if ret:
        # This is not supported by AWS yet
        instance["enclave_options"] = ret.ret["EnclaveOptions"]["Enabled"]

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="ramdisk", InstanceId=instance_id
    )
    if ret:
        instance["ram_disk_id"] = ret.ret["RamdiskId"]

    ret = await hub.exec.boto3.client.ec2.describe_tags(
        ctx,
        Filters=[
            {"Name": "resource-id", "Values": [instance_id]},
        ],
    )
    if ret:
        instance["tags"] = {tag["Key"]: tag.get("Value") for tag in ret.ret["Tags"]}

    # TODO these are available in launch_config and need to be collected manually too
    #    security_group_ids = launch_config["SecurityGroupIds"]
    #    instance_market_options = launch_config["InstanceMarketOptions"]
    #    credit_specification = launch_config["CreditSpecification"]
    #    cpu_options = launch_config["CpuOptions"]
    #    capacity_reservation_specification = launch_config["CapacityReservationSpecification"]
    #    license_specifications = launch_config["LicenseSpecifications"]
    #    hibernation_options = launch_config["HibernationOptions"]
    #    metadata_options = launch_config["MetadataOptions"]
    #    instance_requirements = launch_config["InstanceRequirements"]
    #    private_dns_name_options = launch_config["PrivateDnsNameOptions"]
    #    maintentance_options = launch_config["MaintenanceOptions"]
    return instance


async def get(
    hub, ctx, *, instance_id: str = None, client_token: str = None
) -> Dict[str, Any]:
    """
    Get the state of a single instance from the conversion to present

    This is the preferred "meta" function for collecting information about a single instance
    """
    if instance_id:
        ret = await hub.exec.boto3.client.ec2.describe_instances(
            ctx, InstanceIds=[instance_id]
        )
    elif client_token:
        ret = await hub.exec.boto3.client.ec2.describe_instances(
            ctx, Filters=[dict(Name="client-token", Values=[client_token])]
        )
    else:
        raise ValueError("Must specify either 'instance_id' or 'client_token'")

    if not ret.result:
        return {}

    present_states = await hub.tool.aws.ec2.instance.state.convert_to_present(
        ctx, ret.ret
    )
    if not present_states:
        return {}

    # There will only be one result from "convert_to_present", return it as a plain dictionary
    state = next(iter((present_states).values()))

    if "old_state" in ctx:
        # Check for completed bootstrap calls, this can change when heist can detect existing services
        # Unlike all the other attributes of a state, this can only come from ctx for now
        state["bootstrap"] = (ctx.old_state or {}).get("bootstrap", [])
    else:
        state["bootstrap"] = []

    return state


def test(hub, **kwargs) -> Dict[str, Any]:
    """
    Compute the state based on the parameters passed to an instance.present function for ctx.test
    """
    result = {}
    for k, v in kwargs.items():
        # Ignore kwargs that were None
        if v is None:
            continue
        result[k] = v
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_rest_api_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem apigateway.rest_api present state

     Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert
        idem_resource_name (Text, Optional): An idem name of the resource

    Returns: Valid idem state for apigateway.rest_api of type Dict['string', Any]
    """
    resource_id = raw_resource.get("id")
    resource_parameters = OrderedDict(
        {
            "description": "description",
            "version": "version",
            "binaryMediaTypes": "binary_media_types",
            "minimumCompressionSize": "minimum_compression_size",
            "apiKeySource": "api_key_source",
            "endpointConfiguration": "endpoint_configuration",
            "policy": "policy",
            "tags": "tags",
            "disableExecuteApiEndpoint": "disable_execute_api_endpoint",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



import copy
from typing import Any
from typing import Dict

"""
Exec functions for AWS API Gateway v2 tags.
"""


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    r"""
    Updates tags for an AWS API Gateway v2 resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_arn: AWS API Gateway v2 resource arn.
        old_tags(Dict[str, Any]): List of existing tags.
        new_tags(Dict[str, Any]): List of new tags.

    Returns:
        Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret=None)

    tags_to_add = {}
    tags_to_remove = []
    tags_result = copy.deepcopy(old_tags)

    for key, value in new_tags.items():
        if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
            key not in old_tags
        ):
            tags_to_add[key] = value

    for key in old_tags:
        if key not in new_tags:
            tags_to_remove.append(key)

    if tags_to_remove:
        if not ctx.get("test", False):
            untag_ret = await hub.exec.boto3.client.apigatewayv2.untag_resource(
                ctx, ResourceArn=resource_arn, TagKeys=tags_to_remove
            )
            if not untag_ret["result"]:
                result["result"] = False
                result["comment"] = untag_ret["comment"]
                return result
        [tags_result.pop(key, None) for key in tags_to_remove]

    if tags_to_add:
        if not ctx.get("test", False):
            tag_ret = await hub.exec.boto3.client.apigatewayv2.tag_resource(
                ctx, ResourceArn=resource_arn, Tags=tags_to_add
            )
            if not tag_ret["result"]:
                result["result"] = False
                result["comment"] = tag_ret["comment"]
                return result

    if ctx.get("test", False):
        result["comment"] = (
            f"Would update tags: Add {list(tags_to_add.keys())} Remove {tags_to_remove}",
        )
    else:
        result["comment"] = (
            f"Updated tags: Add {list(tags_to_add.keys())} Remove {tags_to_remove}",
        )
    result["ret"] = {"tags": {**tags_result, **tags_to_add}}
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict


async def search(hub, ctx, name) -> Dict[str, Any]:
    """
    Details about the IAM user or role whose credentials are used to call the operation.

    Args:
        name(string): An Idem name of the resource.

    Request Syntax:
        [Idem-resource-state-name]:
          aws.caller_identity.search:
          - name: 'string'

    Examples:

            my-caller-account:
              aws.caller_identity.search:
                - name: value

    Response Syntax:
          {
            name: 'string',
            user_id: 'string',
            account_id: 'string',
            arn: 'string'
          }

    Response Structure:
        name(string): An Idem name of the resource.
        user_id(string): The unique identifier of the calling entity.
        account_id(string): AWS account ID number of the account that owns or contains the calling entity.
        arn(string): AWS ARN associated with the calling entity.

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    ret = await hub.exec.boto3.client.sts.get_caller_identity(ctx)
    if not ret["result"]:
        result["result"] = False
        result["comment"] = ret["comment"]
        return result
    resource_parameters = OrderedDict(
        {"UserId": "user_id", "Account": "account_id", "Arn": "arn"}
    )
    resource_translated = {"name": name}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in ret["ret"]:
            resource_translated[parameter_present] = ret["ret"][parameter_raw]
    result["old_state"] = resource_translated
    result["new_state"] = copy.deepcopy(result["old_state"])
    return result



from typing import Any
from typing import Dict


async def get_all_hosted_zones(hub, ctx) -> Dict[str, Any]:
    """
    Describes all the hosted zones available.

    Args:
        hub:
        ctx:

    Returns:
        {"result": True|False, "comment": "A message Tuple", "ret": None}
    """
    result = dict(comment=(), result=True, ret=None)
    hosted_zones = []
    ret = await hub.exec.boto3.client.route53.list_hosted_zones(ctx)
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] = ret["comment"]
        return result
    for hosted_zone in ret["ret"]["HostedZones"]:
        resource_id = hosted_zone.get("Id")
        ret = await hub.exec.boto3.client.route53.get_hosted_zone(ctx, Id=resource_id)
        temp_hosted_zone = ret["ret"]["HostedZone"]

        if ret["ret"].get("VPCs"):
            temp_hosted_zone["VPCs"] = ret["ret"].get("VPCs")
        if ret["ret"].get("DelegationSet"):
            temp_hosted_zone["DelegationSet"] = ret["ret"].get("DelegationSet")

        tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
            ctx, ResourceType="hostedzone", ResourceId=resource_id.split("/")[-1]
        )
        if tags_ret["result"] and tags_ret["ret"].get("ResourceTagSet").get("Tags"):
            temp_hosted_zone["Tags"] = tags_ret["ret"]["ResourceTagSet"].get("Tags")
        hosted_zones.append(temp_hosted_zone)

    result["ret"] = {"HostedZones": hosted_zones}
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.create_policy
hub.exec.boto3.client.iam.delete_policy
hub.exec.boto3.client.iam.get_policy
hub.exec.boto3.client.iam.list_policies
hub.exec.boto3.client.iam.tag_policy
hub.exec.boto3.client.iam.untag_policy
resource = hub.tool.boto3.resource.create(ctx, "iam", "Policy", name)
hub.tool.boto3.resource.exec(resource, attach_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_role, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_user, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_version, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_role, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_user, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    policy_document: Dict or str,
    resource_id: str = None,
    path: str = "/",
    description: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Create or update a managed policy for your Amazon Web Services account. When creating a policy, this operation
     creates a policy version with a version identifier of v1 and sets v1 as the policy's default version. When updating
      a policy, this operation creates a new policy version, sets the new policy to the default version, and
       deletes the old policy. As a best practice, you can validate your IAM policies. To learn more, see Validating
        IAM policies in AWS's IAM User Guide. For more information about managed policies in general,
         see Managed policies and inline policies in AWS's IAM User Guide.

    Args:
        name(Text): IAM policy name.
        policy_document(Dict or Text): The JSON policy document that you want to use as the content for the new policy.
         You must provide policies in JSON format in IAM. However, for CloudFormation templates formatted in YAML,
          you can provide the policy in JSON or YAML format. CloudFormation always converts a YAML policy to JSON format
           before submitting it to IAM.
        resource_id(Text, optional): Amazon Resource Name (ARN) to identify IAM policy.
         This is a required field for Idem to manage a brown-field resource.
        path(Text, optional): The path for the policy. Defaults to a slash (/)
        description(Text, optional): A friendly description of the policy.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the IAM customer managed policy.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key): The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.
            * (Value): The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS IAM Policy.
            * create (Dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(Dict, optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [policy-idem-state]:
          aws.iam.policy.present:
          - name: 'string'
          - resource_id: 'string'
          - path: 'string'
          - policy_document: 'dict' or 'string'
          - description: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'
          - timeout:
              create:
                delay: 'int'
                max_attempts: 'int'
              update:
                delay: 'int'
                max_attempts: 'int'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-iam-policy:
              aws.iam.policy.present:
                - name: idem-policy-automation
                - resource_id: arn:aws:iam::53721234567:policy/idem-policy-automation
                - path: /
                - policy_document: '{"Statement": [{"Action": ["ec2:CreateSubnet"], "Effect": "Allow", "Resource": "*"}], "Version": "2012-10-17"}'
                - description: "Idem IAM policy description example"
                - tags:
                  - Key: Name
                    Value: Idem-test-policy
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None

    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )

    if resource_id:
        resource = hub.tool.boto3.resource.create(ctx, "iam", "Policy", resource_id)
        before = await hub.tool.boto3.resource.describe(resource)

    # Standardise on the json format
    policy_document = hub.tool.aws.state_comparison_utils.standardise_json(
        policy_document
    )

    update_ret = None
    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
            ctx, before
        )
        result["comment"] = (f"'{name}' already exists",)
        plan_state = copy.deepcopy(result["old_state"])
        # Update policy document if needed
        if (
            result["old_state"] is not None
            and result["old_state"]["policy_document"]
            and not hub.tool.aws.state_comparison_utils.is_json_identical(
                result["old_state"]["policy_document"], policy_document
            )
        ):
            if ctx.get("test", False):
                plan_state["policy_document"] = policy_document
                result["comment"] = result["comment"] + (
                    f"Would update policy document.",
                )
            else:
                update_ret = await hub.exec.aws.iam.policy.update_policy(
                    ctx=ctx,
                    policy_arn=resource_id,
                    policy_version_id=result["old_state"].get("default_version_id"),
                    new_policy_document=policy_document,
                    timeout=timeout,
                )
                if not update_ret["result"]:
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = False

        if (
            result["old_state"] is not None
            and tags is not None
            and tags != result["old_state"].get("tags")
        ):
            update_ret = await hub.exec.aws.iam.policy.update_policy_tags(
                ctx=ctx,
                police_arn=resource_id,
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]
                result["comment"] = result["comment"] + ("Would update tags.",)

        if update_ret is not None and result["result"] and not ctx.get("test", False):
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.iam.policy", name=name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": f"policy-{name}-resource_id",
                    "id": f"policy-{name}-id",
                    "path": path,
                    "policy_document": policy_document,
                    "description": description,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.iam.policy", name=name
            )
            return result

        ret = await hub.exec.boto3.client.iam.create_policy(
            ctx,
            PolicyName=name,
            Path=path,
            PolicyDocument=policy_document,
            Description=description,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=1,
            default_max_attempts=40,
            timeout_config=timeout.get("create") if timeout else None,
        )
        hub.log.debug(f"Waiting on creation of aws.iam.policy '{name}'")
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "iam",
                "policy_exists",
                PolicyArn=ret["ret"].get("Policy")["Arn"],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.iam.policy", name=name
        )
        resource_id = ret["ret"]["Policy"]["Arn"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif before and (update_ret is None):
            result["new_state"] = copy.deepcopy(result["old_state"])
        else:
            resource = hub.tool.boto3.resource.create(ctx, "iam", "Policy", resource_id)
            after = await hub.tool.boto3.resource.describe(resource)
            result[
                "new_state"
            ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, after
            )
    except Exception as e:
        result["comment"] = (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified managed policy. Before you can delete a managed policy, you must first detach the policy
    from all users, groups, and roles that it is attached to. In addition, you must delete all the policy's
    versions. The following steps describe the process for deleting a managed policy:   Detach the policy from all
    users, groups, and roles that the policy is attached to, using DetachUserPolicy, DetachGroupPolicy, or
    DetachRolePolicy. To list all the users, groups, and roles that a policy is attached to, use
    ListEntitiesForPolicy.   Delete all versions of the policy using DeletePolicyVersion. To list the policy's
    versions, use ListPolicyVersions. You cannot use DeletePolicyVersion to delete the version that is marked as the
    default version. You delete the policy's default version in the next step of the process.   Delete the policy
    (this automatically deletes the policy's default version) using this operation.   For information about managed
    policies, see Managed policies and inline policies in the IAM User Guide.

    Args:
        name(Text): IAM policy name.
        resource_id(Text, optional): Amazon Resource Name (ARN) to identify IAM policy. If not specified, Idem will use "name"
         parameter to identify the IAM policy on AWS.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.iam.policy.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.policy", name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "iam", "Policy", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.policy", name=name
        )
    else:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, raw_resource=before
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.iam.policy", name=name
            )
            return result

        ret = await hub.exec.boto3.client.iam.delete_policy(
            ctx, PolicyArn=resource_id if resource_id else name
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.iam.policy", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists all of your own customer-defined managed policies.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.policy
    """

    result = {}
    # Set scope to local to only list the customer-defined managed policies.
    ret_policies = await hub.exec.boto3.client.iam.list_policies(ctx, Scope="Local")

    if not ret_policies["result"]:
        hub.log.debug(f"Could not describe policy {ret_policies['comment']}")
        return {}

    for policy in ret_policies["ret"]["Policies"]:
        translated_resource = (
            await hub.tool.aws.iam.conversion_utils.convert_raw_policy_to_present(
                ctx, policy
            )
        )
        if translated_resource is None:
            continue
        resource_key = f"iam-policy-{translated_resource['resource_id']}"
        result[resource_key] = {
            "aws.iam.policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions for AWS API Gateway v2 Integration resources.
"""


def convert_raw_integration_to_present(
    hub, api_id: str, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway v2 Integration resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        api_id(string): The API resource identifier in Amazon Web Services.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(string, optional): An Idem name of the resource.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_parameters = OrderedDict(
        {
            "ApiGatewayManaged": "api_gateway_managed",
            "ConnectionId": "connection_id",
            "ConnectionType": "connection_type",
            "ContentHandlingStrategy": "content_handling_strategy",
            "CredentialsArn": "credentials_arn",
            "Description": "description",
            "IntegrationMethod": "integration_method",
            "IntegrationResponseSelectionExpression": "integration_response_selection_expression",
            "IntegrationSubtype": "integration_subtype",
            "IntegrationType": "integration_type",
            "IntegrationUri": "integration_uri",
            "PassthroughBehavior": "passthrough_behavior",
            "PayloadFormatVersion": "payload_format_version",
            "RequestParameters": "request_parameters",
            "RequestTemplates": "request_templates",
            "ResponseParameters": "response_parameters",
            "TemplateSelectionExpression": "template_selection_expression",
            "TimeoutInMillis": "timeout_in_millis",
            "TlsConfig": "tls_config",
        }
    )
    resource_translated = {
        "name": idem_resource_name
        if idem_resource_name
        else raw_resource.get("IntegrationId"),
        "resource_id": raw_resource.get("IntegrationId"),
        "api_id": api_id,
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



import pathlib

from Cryptodome.PublicKey import RSA


async def bootstrap(
    hub,
    ctx,
    instance_id,
    username: str = None,
    host: str = None,
    ssh_public_key: str = None,
    ssh_private_key: str = None,
    availability_zone: str = None,
    *,
    heist_manager: str,
    artifact_version: str = None,
    **kwargs,
):
    """
    Connect to the instance with an ssh keypair and call heist to bootstrap it.
    If an ssh keypair is not provided one will be created.

    Args:
        hub:
        ctx:
        instance_id: An AWS EC2 Instance ID
        username(Text, Optional): The instance OS username to use in the connection, Defaults to ec2-user
        host(Text, Optional): The public ip address or dns name of the instance.  Defaults to autodetect
        ssh_public_key(Text, Optional): A public ssh key or path to send to the instance
        ssh_private_key(Text, Optional): A private ssh key or path to send to the instance
        availability_zone(Text, Optional): The Availability Zone in which the EC2 instance was launched
        heist_manager(Text, Required): The heist manager to use to bootstrap the instance (I.E. salt.minion)
        artifact_version(Text, Optional): The version of the heist_manager's artifact to upload to the instance, Defaults to latest

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    Examples:

        .. code-block:: bash

            idem exec aws.ec2.instance.bootstrap <instance_id> heist_manager="salt.minion"

    """
    result = dict(comment=(), result=True, ret=None)

    if bool(ssh_public_key) is not bool(ssh_private_key):
        result["comment"] += (
            "Both a private ssh key and a public ssh key must be specified",
        )
        result["result"] = False
        return result

    if not (ssh_public_key and ssh_private_key):
        result["comment"] += (
            f"Created a keypair to bootstrap instance with {heist_manager}",
        )
        key_pair = RSA.generate(2048)
        ssh_public_key = key_pair.publickey().export_key("OpenSSH")
        ssh_private_key = key_pair.export_key("PEM")

    # Create an instance resource object to use for defaults
    instance = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", instance_id)

    # Verify the public key
    public_key_path = pathlib.Path(ssh_public_key)
    # If the ssh_public_key was a path, then read the contents of that path into a variable
    if public_key_path.exists():
        ssh_public_key = public_key_path.read_text()

    # Verify the private key
    private_key_path = pathlib.Path(ssh_private_key)
    if private_key_path.exists():
        ssh_private_key = private_key_path.read_bytes()
    else:
        ssh_private_key = ssh_private_key.encode()

    # Verify the host address
    if host is None:

        # Default to the public ip address and fall back to the public dns name
        host = instance.public_ip_address or instance.public_dns_name

    # Verify the username
    if username is None:
        # This is a safe bet for most official AMIs except Windows, Debian, Ubuntu, and Bitnami
        username = "ec2-user"
        # TODO, can we do more intelligent username guessing based on the AMI?
        #     https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/managing-users.html
        #     It can probably come from UserData on the instance...

    # Verify the availability zone
    if availability_zone is None:
        availability_zone = instance.placement["AvailabilityZone"]

    # Connect to the ec2 instance with the public key, connection will be available for 60 seconds
    connection_ret = (
        await hub.exec.boto3.client.ec2_instance_connect.send_ssh_public_key(
            ctx,
            InstanceId=instance_id,
            InstanceOSUser=username,
            SSHPublicKey=ssh_public_key,
            AvailabilityZone=availability_zone,
        )
    )

    result["result"] &= connection_ret.result & connection_ret.ret["Success"]
    if not result["result"]:
        result["comment"] += (
            f"Failed to connect to instance {instance_id}",
            f"InstanceConnect RequestId: {connection_ret.ret['RequestId']}",
            connection_ret.comment,
        )
        return result

    remotes = {
        instance_id: dict(
            host=host,
            username=username,
            client_keys=[ssh_private_key],
            **kwargs,
        ),
        "bootstrap": True,
    }

    await hub.heist[heist_manager].run(
        remotes=remotes, artifact_version=artifact_version
    )

    result["comment"] += (
        f"Successfully bootstrapped instance '{instance_id}' with '{heist_manager}'",
    )

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions for AWS Lambda Function resources.
"""


async def convert_raw_lambda_function_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    Converts an AWS Lambda Function resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(string, optional): An Idem name of the resource.
        tags(List, optional): The AWS Lambda Function tags.

    Returns:
        Dict[str, Any]: Common idem present state.
    """
    describe_parameters = OrderedDict(
        {
            "FunctionArn": "function_arn",
            "Runtime": "runtime",
            "Role": "role",
            "Handler": "handler",
            "Description": "description",
            "Timeout": "function_timeout",
            "MemorySize": "memory_size",
            "VpcConfig": "vpc_config",
            "PackageType": "package_type",
            "DeadLetterConfig": "dead_letter_config",
            "Environment": "environment",
            "KMSKeyArn": "kms_key_arn",
            "TracingConfig": "tracing_config",
            "Layers": "layers",
            "FileSystemConfigs": "file_system_configs",
            "Architectures": "architectures",
        }
    )
    new_function = {"resource_id": idem_resource_name, "name": idem_resource_name}
    result = dict(comment=(), result=True, ret=None)
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if raw_resource.get(parameter_old_key) is not None:
            new_function.update(
                {parameter_new_key: raw_resource.get(parameter_old_key)}
            )
    config_arn = await hub.exec.boto3.client["lambda"].get_function_code_signing_config(
        ctx, FunctionName=raw_resource.get("FunctionArn")
    )
    result["result"] = config_arn["result"]
    if not result["result"]:
        result["comment"] = result["comment"] + config_arn["comment"]

    if config_arn.get("result") and config_arn.get("ret").get("CodeSigningConfigArn"):
        new_function.update(
            {
                "code_signing_config_arn": config_arn.get("ret").get(
                    "CodeSigningConfigArn"
                )
            }
        )

    if raw_resource.get("ImageConfigResponse") and raw_resource.get(
        "ImageConfigResponse"
    ).get("ImageConfig"):
        new_function.update(
            {
                "image_config": (
                    raw_resource.get("ImageConfigResponse").get("ImageConfig")
                ).copy()
            }
        )
    if tags:
        new_function["tags"] = tags.copy()

    result["ret"] = new_function
    return result
